from contextlib import contextmanager
import logging
import pathlib
from typing import Optional, Union

import retinapy._logging
import retinapy.models
import torch


_logger = logging.getLogger(__name__)

"""
Max training time before a new recovery is made.

Standard checkpointing only happens every epoch. But a checkpoint for recovery
purposes will be made every 30 minutes. Only a single recovery is kept.
"""
RECOVERY_CKPT_PERIOD_SEC = 30 * 60

"""
Here we take part in the rite of of passage for a deep learning project by
yet again reinventing the training loop architecture. No one wants their 
project stitched together with the callbacks of some soon to be abandonded or 
rewritten DL framework.
"""


class Trainable:
    """Encapsulates a dataset, model input-output and loss function.

    This class is needed in order to be able to train multiple models and
    configurations with the same training function. The training function
    is too general to know about how to route the data into and out of a model,
    evaluate the model or how to take a model output and create a prediction.

    Redesign from function parameters to a class
    --------------------------------------------
    The class began as a dumb grouping of the parameters to the train function:
    train_ds, val_ds, test_ds, model, loss_fn, forward_fn, val_fn, and more—there
    were so many parameters that they were grouped together into a NamedTuple.
    However, functions like forward_fn and val_fn would need to be given
    the model, loss_fn and forward_fn in order to operate. This is the exact
    encapsulation behaviour that classes achieve, so the NamedTuple was made
    into a class. Leaning into the use of classes, forward_fn and val_fn were
    made into methods of the class, while the rest became properties. This
    change is noteworthy as customizing the forward or validation functions
    now requires defining a new class, rather than simply passing in a new
    function. Although, nothing is stopping someone from defining a class that
    simply wraps a function and passes the arguments through.

    Flexibility to vary the data format
    -----------------------------------
    Why things belongs inside or outside this class can be understood by
    realizing that the nature of the datasets are encapsulated here. As such,
    any function that needs to extract the individual parts of a dataset
    sample will need to know what is in each sample. Such a function is a
    good candidate to appear in this class.

    While in some cases you can separate the datasets from the models, this
    isn't always easy or a good idea. A model for ImageNet can easily be
    separated from the dataset, as the inputs and outputs are so standard; but
    for the spike prediction, the model output is quite variable. Consider
    the distance-field model which outputs a distance field, whereas a
    Poisson-distribution model will output a single number. Training is
    done with these outputs, and involves the dataset producing sample tuples
    that have appropriate elements (distance fields, for example).
    The actual inference is an additional calculation using these outputs.

    So, the procedure of taking the model input and model output from a dataset
    sample and feeding it to the model calculating the loss and doing the
    inference—none of these steps can be abstracted to be ignorant of either
    the model or the dataset.

    Other libraries
    ---------------
    Compared to Keras and FastAI: Trainable encapsulates a lot less than
    Keras's Model or FastAI's Learner.

    At this point (2022-09-12) I'm not eager to use the FastAI API, as I don't
    want to discover later that it's too limiting in some certain way. It's
    quite possible that it's already too prescriptive. Reading the docs, it's
    not clear what parts of Learner's internals are exposed for customization.
    If all "_" prefixed methods are not meant to be customized, then it's
    already too restrictive. Notably, there seems to be an expected format for
    the elements of the dataset, which I want to avoid. The reason for this is
    that the distance fields are intermediate results, and while I want to
    train on them, I would like to evaluate based on approximations to
    actual spike count accuracy, and I would like to make predictions using
    much more expensive dynamic programming inference routines. So the data
    doesn't fall nicely into (X,y) type data, and the metrics are not
    consistent across training and evaluation.

    In addition, at least at the momemt, FastAI's library provides a lot more
    abstraction/generalization than I need, which can make it harder for
    myself (or others) to understand what is going on. This might end up being
    a mistake, as the growing code might reveal itself to provide nice
    abstraction boundaries that are already handled nicely in FastAI.
    """

    def __init__(
        self,
        train_ds: torch.utils.data.Dataset,
        val_ds: torch.utils.data.Dataset,
        test_ds: torch.utils.data.Dataset,
        model: torch.nn.Module,
        label: str,
        # TODO: manage GPU usage.
    ):
        """
        Args:
            train_ds: the dataset to train the model on.
            val_ds: the dataset to evaluate the model on and guide model
                training. This dataset is used to decide what model states to
                keep, and when to stop training, if early termination is
                enabled.
            test_ds: the test dataset. Similar to the validation dataset, this
                dataset is available for evaluating the model; however,
                its purpose is to be a datasat which has no influence
                on guiding the training. This includes any hyperparameter
                turing and the design of inference procedures. If more stages
                of data holdout are desired, then the validation dataset
                should be split again, rather than using the test dataset.
            model: the PyTorch model to train.
            label: a string label for this trainable.
        """
        self.train_ds = train_ds
        self.val_ds = val_ds
        self.test_ds = test_ds
        self.model = model
        self.label = label

    def forward(self, sample):
        """Run the model forward.

        Args:
            sample: a single draw from the train or validation data loader.

        Returns:
            (output, loss): the model output and the loss, as a tuple.
        """
        raise NotImplementedError("Override")

    def evaluate(self, val_dl):
        """Run the full evaluation procedure.

        Args:
            val_dl: the validation data loader.

        Returns:
            metrics: a str:float dictionary containing evaluation metrics. It
                is expected that this dictionary at least contains 'loss' and
                'accuracy' metrics.
        """
        raise NotImplementedError("Override")

    def in_device(self):
        """Returns the device on which the model expects input to be located.

        Most likely, the whole model is on a single device, and it is
        sufficient to use `next(self.model.parameters()).device`.
        """
        raise NotImplementedError("Override")

    def __str__(self) -> str:
        return f"Trainable ({self.label})"

    def model_summary(self, sample) -> str:
        """Returns a detailed description of the model.

        Args:
            sample: a training sample given to allow the model structure to
            be inferred.

        At the moment, this is called by train() with the intent of saving
        out a file containing info like torchinfo.summary. The torch module
        input shape isn't known by train(), so the actual summary creation
        must be done somewhere like Trainable.

        Override this to add more features.
        """
        return f"Trainable ({self.label})"


def _create_dataloaders(train_ds, val_ds, test_ds, batch_size, 
                        num_workers, pin_memory):
    # Setting pin_memory=True. This is generally recommended when training on
    # Nvidia GPUs. See:
    #   - https://discuss.pytorch.org/t/when-to-set-pin-memory-to-true/19723
    #   - https://developer.nvidia.com/blog/how-optimize-data-transfers-cuda-cc/
    train_dl = torch.utils.data.DataLoader(
        train_ds,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )
    val_dl = torch.utils.data.DataLoader(
        val_ds,
        batch_size=batch_size,
        # For debugging, it's nice to see a variety:
        shuffle=True,
        drop_last=True,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )
    test_dl = torch.utils.data.DataLoader(
        test_ds,
        batch_size=batch_size,
        shuffle=False,
        drop_last=False,
        num_workers=num_workers,
    )
    return train_dl, val_dl, test_dl


@contextmanager
def evaluating(model):
    """
    Context manager to set the model to eval mode and then back to train mode.

    Used this to prevent an exception leading to unexpected training state.
    """
    original_mode = model.training
    model.eval()
    try:
        model.eval()
        yield
    finally:
        # Switch back to the original training mode.
        model.train(original_mode)


class TrainingTimers:
    """Collect timers here for convenience."""

    def __init__(self):
        self.batch = retinapy._logging.Timer()
        self.epoch = retinapy._logging.Timer()
        self.validation = retinapy._logging.Timer()
        self.recovery = retinapy._logging.Timer()

    @staticmethod
    def create_and_start():
        timer = TrainingTimers()
        timer.batch.restart()
        timer.epoch.restart()
        timer.recovery.restart()
        return timer


def train(
    trainable: Trainable,
    num_epochs: int,
    batch_size: int,
    lr: float,
    weight_decay: float,
    out_dir: Union[str, pathlib.Path],
    steps_til_log: int = 1000,
    steps_til_eval: Optional[int] = None,
    evals_til_eval_test_ds: Optional[int] = None,
    initial_checkpoint: Optional[Union[str, pathlib.Path]] = None,
    num_workers: int = 4,
    pin_memory: bool = False,
):
    """
    Train a model.

    This is a training loop that works with any Trainable object.

    It encapsulates basic functionality like logging, checkpointing and
    choosing when to run an evalutaion. Users might be just as well off
    by copying the code to use as a baseline and modifying it to their needs.
    """
    logging.info(f"Training {trainable.label}")

    # Enable Cuda convolution auto-tuning. For more details, see:
    #   https://pytorch.org/tutorials/recipes/recipes/tuning_guide.html
    torch.backends.cudnn.benchmark = True

    # Create a GradScalar. For more details, see:
    #   https://pytorch.org/docs/stable/notes/amp_examples.html#amp-examples
    grad_scaler = torch.cuda.amp.GradScaler()

    # Setup output (logging & checkpoints).
    tensorboard_dir = pathlib.Path(out_dir) / "tensorboard"
    tb_logger = retinapy._logging.TbLogger(tensorboard_dir)


    # Load the model & loss fn.
    # The order here is important when resuming from checkpoints. We must:
    # 1. Create model.
    # 2. Send model to target device.
    # 3. Create optimizer.
    # 4. Initialize from checkpoint.
    #
    # The reason the order is crucial is that the optimizer must be on the gpu
    # before having it's parameters populated, as there is no optimizer.gpu()
    # method (possibly coming: https://github.com/pytorch/pytorch/issues/41839).
    # An alternative would be to use the map_location argument. See the 
    # discussion: https://github.com/pytorch/pytorch/issues/2830.
    model = trainable.model
    model.cuda()
    optimizer = torch.optim.AdamW(
        model.parameters(), lr=lr, weight_decay=weight_decay
    )
    if initial_checkpoint is not None:
        retinapy.models.load_model_and_optimizer(model, optimizer, 
                                                 initial_checkpoint)

    # Load the data.
    train_dl, val_dl, test_dl = _create_dataloaders(
        trainable.train_ds,
        trainable.val_ds,
        trainable.test_ds,
        batch_size=batch_size,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )
    _logger.info(
        f"Dataset sizes: train ({len(train_dl.dataset)}), "
        f"val ({len(val_dl.dataset)}), test ({len(test_dl.dataset)})."
    )
    model.train()
    # Before going any further, log model structure.
    out_file = pathlib.Path(out_dir) / "model_summary.txt"
    with open(out_file, "w") as f:
        # This is allowed to fail, as if the model has issues, we want to
        # get the errors from the actual training forward pass.
        summary = None
        try:
            summary = trainable.model_summary(next(iter(train_dl)))
        except Exception as e:
            msg = (
                "Failed to generating model summary. Exception raised:\n"
                f"{str(e)}"
            )
            _logger.error(msg)
            summary = msg
        f.write(str(summary))

    model_saver = retinapy._logging.ModelSaver(out_dir, model, optimizer)
    metric_tracker = retinapy._logging.MetricTracker(out_dir)

    def _eval():
        nonlocal num_evals
        if num_evals == evals_til_eval_test_ds:
            dl = test_dl
            label = "test-ds"
        else:
            dl = val_dl
            label = "val-ds"
        _logger.info(f"Running evaluation {label}")
        with evaluating(trainable.model), torch.no_grad(), timers.validation:
            eval_results = trainable.evaluate(dl)
            tb_logger.log(step, eval_results, label)
            if "metrics" not in eval_results:
                raise ValueError("Trainable.evaluate() must return metrics.")
            metrics = eval_results["metrics"]
            retinapy._logging.print_metrics(metrics)
        tb_logger.log_scalar(
            step, "eval-time", timers.validation.elapsed(), label
        )
        num_evals += 1
        _logger.info(
            f"Finished evaluation in {round(timers.validation.elapsed())} sec "
            f"(rolling ave: {round(timers.validation.rolling_duration())} sec)"
        )
        num_evals += 1
        return metrics

    _logger.info("Starting training loop.")
    step = 0
    num_evals = 0
    timers = TrainingTimers.create_and_start()
    for epoch in range(num_epochs):
        timers.epoch.restart()
        loss_meter = retinapy._logging.Meter("loss")
        for batch_step, sample in enumerate(train_dl):
            timers.batch.restart()
            # set_to_none=True is suggested to improve performance, according to:
            #   https://pytorch.org/tutorials/recipes/recipes/tuning_guide.html
            # A recipe for autocast, grad scaling and set_to_none:
            #   https://pytorch.org/tutorials/recipes/recipes/amp_recipe.html
            optimizer.zero_grad(set_to_none=True)
            with torch.cuda.amp.autocast():
                model_out, total_loss = trainable.forward(sample)
            grad_scaler.scale(total_loss).backward()
            grad_scaler.step(optimizer)
            grad_scaler.update()

            # Loss is expected to be returned already batch averaged.
            loss_meter.update(total_loss.item())
            metrics = [
                retinapy._logging.Metric("loss", total_loss.item()),
            ]
            tb_logger.log_metrics(step, metrics, log_group="train")

            # Log to console.
            if step % steps_til_log == 0:
                model_mean = torch.mean(model_out)
                model_sd = torch.std(model_out)
                elapsed_min, elapsed_sec = divmod(
                    round(timers.epoch.elapsed()), 60
                )
                # Floor total minutes to align with batch minutes.
                total_hrs, total_min = divmod(
                    int(timers.epoch.total_elapsed() / 60), 60
                )
                _logger.info(
                    f"epoch: {epoch}/{num_epochs} | "
                    f"step: {batch_step:>4}/{len(train_dl)} "
                    f"({batch_step/len(train_dl):>3.0%}) "
                    f"{round(1/timers.batch.rolling_duration()):>2}/s | "
                    f"elapsed: {elapsed_min:>1}m:{elapsed_sec:02d}s "
                    f"({total_hrs:>1}h:{total_min:02d}m) | "
                    f"loss: {loss_meter.avg:.3f} | "
                    f"out μ (σ): {model_mean:>3.2f} ({model_sd:>3.2f})"
                )
                loss_meter.reset()

            # Evaluate.
            # (step + 1), as we don't want to evaluate on the first step.
            if steps_til_eval and (batch_step + 1) % steps_til_eval == 0:
                is_near_epoch_end = batch_step + steps_til_eval >= len(train_dl)
                if not is_near_epoch_end:
                    _eval()
            step += 1

            # Recovery.
            # Don't allow training to proceed too long without checkpointing.
            if timers.recovery.elapsed() > RECOVERY_CKPT_PERIOD_SEC:
                model_saver.save_recovery()
                timers.recovery.restart()

        _logger.info(
            f"Finished epoch in {round(timers.epoch.elapsed())} secs "
            f"(rolling duration: "
            f"{round(timers.epoch.rolling_duration())} s/epoch)"
        )
        # Evaluate and save at end of epoch.
        metrics = _eval()
        # If this on_metric_end type of behaviour grows, consider switching
        # to callbacks.
        improved_metrics = metric_tracker.on_epoch_end(metrics, epoch)
        model_saver.save_checkpoint(epoch, improved_metrics)
        timers.recovery.restart()
    _logger.info(
        f"Finished training. {round(timers.epoch.total_elapsed())} "
        "secs elsapsed."
    )



import torch
import torch.nn as nn
from torch.nn import functional as F
import math
import einops


def create_batch_norm(n):
    """
    Following Davis and Frank's recommendation in "Revisiting Batch
    Normazilation", we would initialize batch norm weights to less than 1
    (they suggested to use 0.1). They also reccomended using a lower learning
    rate for the γ parameter.

    For comparison, fastai initialize β to 0.001 and γ to 1.

    I tried both, and found better results with fastai's defaults.
    """
    bn = nn.BatchNorm1d(n)
    # fastai
    # bn.weight.data.fill_(1.0)
    # bn.bias.data.fill_(1e-3)
    # Davis and Frank
    bn.weight.data.fill_(0.1)
    bn.bias.data.fill_(0)
    return bn


def create_shortcut(in_n, out_n, stride):
    """Residual connection.

    Currently only "pool" type supported"""
    # 1. Identity
    # In the simplest case, we can just return the input. For this to work,
    # both the channel count and channel dimensions of the input and ouput
    # must match. The output channel dimension is determined by the stride.
    channels_match = in_n == out_n
    downsample = stride > 1
    if channels_match and not downsample:
        return nn.Identity()

    skip_layers = []
    # 2. Downsample
    if downsample:
        # The following (1, 7) input:
        # |  1  |  2  |  3  |  4  |  5  |  6  |  7  |
        # when pooled with stride=kernel=2 becomes (1, 4):
        # |    1.5    |    3.5    |    5.5   | 7 |
        pool = nn.AvgPool1d(
            kernel_size=stride,
            stride=stride,
            count_include_pad=False,
            ceil_mode=True,
        )
        skip_layers.append(pool)
    # 3. 1x1 conv
    if not channels_match:
        # There isn't a consensus on whether:
        #   - to use batch norm or a bias or neither.
        conv = nn.Conv1d(in_n, out_n, kernel_size=1, bias=False)
        skip_layers.append(conv)
    res = nn.Sequential(*skip_layers)
    return res


class Decoder1dBlock(nn.Module):
    """
    I referred to decoder architecture here:
    https://github.com/qubvel/segmentation_models.pytorch/tree/master/segmentation_models_pytorch/decoders
    """

    def __init__(self, in_channels, out_channels, act=True):
        super(Decoder1dBlock, self).__init__()
        self.act = act
        self.conv1 = nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            # No need for bias here, since we're using batch norm.
            # https://pytorch.org/tutorials/recipes/recipes/tuning_guide.html
            bias=False,
        )
        self.bn1 = nn.BatchNorm1d(out_channels)
        self.conv2 = nn.Conv1d(
            out_channels,
            out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            # No need for bias here, since we're using batch norm.
            # https://pytorch.org/tutorials/recipes/recipes/tuning_guide.html
            bias=False,
        )
        self.bn2 = nn.BatchNorm1d(out_channels)

    def forward(self, x):
        x = F.interpolate(x, scale_factor=2, mode="nearest")
        x = F.relu(self.bn1(self.conv1(x)))
        if self.act:
            x = F.relu(self.bn2(self.conv2(x)))
        else:
            x = self.conv2(x)
        return x


class ResBlock1d(nn.Module):
    """A residual block with 1d convolutions.

    This is a pretty inflexible implementation. No need to make it any
    more general yet.
    """

    def __init__(self, in_n, mid_n, out_n, kernel_size=3, downsample=False):
        super(ResBlock1d, self).__init__()
        self.downsample = downsample
        stride = 2 if self.downsample else 1
        self.shortcut = self.create_shortcut(in_n, out_n, stride=stride)
        # Note: bias is False for the conv layers, as they will be followed
        # by batch norm.
        self.conv1 = nn.Conv1d(
            in_n,
            mid_n,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=1,
            bias=False,
        )
        padding = (kernel_size - 1) // 2
        self.conv2 = nn.Conv1d(
            mid_n,
            mid_n,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=1,
            bias=False,
        )
        self.conv3 = nn.Conv1d(
            mid_n,
            out_n,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=1,
            bias=False,
        )
        # To use batch norm or group norm?
        self.bn1 = create_batch_norm(mid_n)
        self.bn2 = create_batch_norm(mid_n)
        self.bn3 = create_batch_norm(out_n)

        self.se = SEModule(out_n)

    @staticmethod
    def create_shortcut(in_n, out_n, stride, downsample_type="pool"):
        """
        The identify path is one of those finicky bits of ResNet type networks.

        Depending on whether the input and output match in terms of channel
        count and dimension, we have the following behaviour:

        Match?
        ------

            | Channel count | Dimensions | Bevaviour                  |
            |---------------|------------|----------------------------|
            |      ✓        |     ✓      | identity                   |
            |      ✓        |     ✘      | pool or conv               |
            |      ✘        |     ✓      | 1x1 conv                   |
            |      ✘        |     ✘      | pool or conv and 1x1 conv  |


        The most interesting choice is whether to use a strided pool or a
        strided convolution to achieve the downsampling effect. It's
        interesting as implementations are split on which one to use. There
        are futher choices too, such as whether to use dilation in addition
        to strides, and whether to downsample before or after the 1x1 conv.

        Some implementations for reference:
            - fastai: https://github.com/fastai/fastai/blob/aa58b1316ad8e7a5fa2e434e15e5fe6df4f4db56/nbs/01_layers.ipynb
            - lightning: https://github.com/Lightning-AI/lightning-bolts/blob/master/pl_bolts/models/self_supervised/resnets.py
            - pytorch image models: https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/resnet.py
                 - there are two functions, downsample_conv() and
                   downsample_avg() that are used to create the downsampling for
                   the shortcut connection.
                 - uses ceil_mode=True, so a dimension n=7 would be reduced to
                   n=4, not n=3.

        My gut instinct is that swapping out pooling for convolution to achieve
        the downsample will naively achieve better results; however, the
        convolution is instrinsically more powerful (actual has paratemers) and
        so, if you care about parameter counts, then a fair comparison would
        involve reducing parameters elsewhere. Given that the whole point of
        the residual layer is to shortcircut a layer and allow gradients to
        flow easily, I think that pooling gets more theory points for staying
        true to this idea. Ironically, I think the original ResNet
        implementation might have used convolution.
        """
        # downsample_types = {"pool", "conv"}
        # Actually, let's just worry about pool for the moment.
        downsample_types = {"pool"}
        if downsample_type not in downsample_types:
            raise ValueError(
                f"Invalid downsample type ({downsample_type}). Expected one "
                f"of ({downsample_types})."
            )
        # 1. Identity
        # In the simplest case, we can just return the input. For this to work,
        # both the channel count and channel dimensions of the input and ouput
        # must match. The output channel dimension is determined by the stride.
        channels_match = in_n == out_n
        downsample = stride > 1
        if channels_match and not downsample:
            return nn.Identity()

        skip_layers = []
        # 2. Downsample
        if downsample:
            # The following (1, 7) input:
            # |  1  |  2  |  3  |  4  |  5  |  6  |  7  |
            # when pooled with stride=kernel=2 becomes (1, 4):
            # |    1.5    |    3.5    |    5.5   | 7 |
            pool = nn.AvgPool1d(
                kernel_size=stride,
                stride=stride,
                count_include_pad=False,
                ceil_mode=True,
            )
            skip_layers.append(pool)
        # 3. 1x1 conv
        if not channels_match:
            # There isn't a consensus on whether:
            #   - to use batch norm or a bias or neither.
            conv = nn.Conv1d(in_n, out_n, kernel_size=1, bias=False)
            skip_layers.append(conv)
        res = nn.Sequential(*skip_layers)
        return res

    def forward(self, x):
        shortcut = self.shortcut(x)
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = self.bn3(self.conv3(x))
        x = self.se(x) + shortcut
        x = F.relu(x)
        return x


class FcLayer(nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(in_features, out_features),
            nn.LayerNorm([out_features]),
            nn.ReLU(inplace=True),
        )

    def forward(self, input):
        return self.net(input)


class FcBlock(nn.Module):
    def __init__(
        self,
        hidden_ch,
        num_hidden_layers,
        in_features,
        out_features,
        outermost_linear=False,
    ):
        super().__init__()

        self.net_elements = []
        self.net_elements.append(nn.Flatten())
        self.net_elements.append(
            FcLayer(in_features=in_features, out_features=hidden_ch)
        )

        for i in range(num_hidden_layers):
            self.net_elements.append(
                FcLayer(in_features=hidden_ch, out_features=hidden_ch)
            )

        if outermost_linear:
            self.net_elements.append(
                nn.Linear(in_features=hidden_ch, out_features=out_features)
            )
        else:
            self.net_elements.append(
                FcLayer(in_features=hidden_ch, out_features=out_features)
            )

        self.net = nn.Sequential(*self.net_elements)
        self.net.apply(self.init_weights)

    def __getitem__(self, item):
        return self.net_elements[item]

    def init_weights(self, m):
        if type(m) == nn.Linear:
            nn.init.kaiming_normal_(
                m.weight, a=0.0, nonlinearity="relu", mode="fan_in"
            )

    def forward(self, input):
        ans = self.net(input)
        return ans


class Conv1dSame(torch.nn.Module):
    """1D convolution that pads to keep spatial dimensions equal.
    Cannot deal with stride. Only quadratic kernels (=scalar kernel_size).
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        bias=True,
        padding_layer=nn.ReflectionPad1d,
    ):
        """
        :param in_channels: Number of input channels
        :param out_channels: Number of output channels
        :param kernel_size: Scalar. Spatial dimensions of kernel (only quadratic kernels supported).
        :param bias: Whether or not to use bias.
        :param padding_layer: Which padding to use. Default is reflection padding.
        """
        super().__init__()
        ka = kernel_size // 2
        kb = ka - 1 if kernel_size % 2 == 0 else ka
        self.net = nn.Sequential(
            padding_layer((ka, kb, ka, kb)),
            nn.Conv1d(
                in_channels, out_channels, kernel_size, bias=bias, stride=1
            ),
        )

        self.weight = self.net[1].weight
        self.bias = self.net[1].bias

    def forward(self, x):
        return self.net(x)


class Conv1dWarehouse(nn.Module):
    def __init__(
        self,
        max_in_channels,
        warehouse_size,
        kernel_size=7,
        key_len=16,
    ):
        super().__init__()
        assert kernel_size == 7, "Currently, only kernel_size=7 is supported."
        self.max_in_channels = max_in_channels
        self.warehouse_size = warehouse_size
        self.key_len = key_len
        self.weights_per_kernel = kernel_size * self.max_in_channels
        self.warehouse_W = nn.Linear(
            self.warehouse_size, self.weights_per_kernel
        )
        self.warehouse_b = nn.Linear(self.warehouse_size, 1)
        self.keys = nn.Linear(self.key_len, self.warehouse_size)

    def forward(self, q):
        """
        q.shape = (batch, num_channels, key_len)
        """
        batch_size = q.shape[0]
        num_kernels = q.shape[1]
        att = F.softmax(self.keys(q), dim=2)
        weights_W = self.warehouse_W(att)
        weights_b = self.warehouse_b(att).unsqueeze(-1)
        return weights_W, weights_b

    def forward_old(self, q, num_in_channels):
        """
        q.shape = (batch, num_channels, key_len)
        """
        batch_size = q.shape[0]
        num_kernels = q.shape[1]
        # soft
        # att = F.softmax(self.keys(q), dim=2)
        # hard
        att = F.argmax(self.keys(q), dim=2, keepdim=True)
        weights_W = self.warehouse_W(att)
        weights_W = weights_W.view(
            batch_size, num_kernels, self.max_in_channels, self.kernel_size
        )
        weights_W = weights_W[:, :, :num_in_channels, :]
        weights_b = self.warehouse_b(att).unsqueeze(-1)
        return weights_W, weights_b


class LayerNorm1d(nn.LayerNorm):
    """LayerNorm for channels of '2D' spatial NCHW tensors"""

    def __init__(self, num_channels, eps=1e-6, affine=True):
        super().__init__(num_channels, eps=eps, elementwise_affine=affine)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.layer_norm(
            x.permute(0, 2, 1),
            self.normalized_shape,
            self.weight,
            self.bias,
            self.eps,
        ).permute(0, 2, 1)


class ResBlock1d_F(nn.Module):
    """Almost functional residual block.

    Need to get rid of bath norm to be fully functional.
    """

    def __init__(
        self,
        in_n,
        mid_n,
        out_n,
        kernel_size=3,
        downsample=False,
    ):
        super().__init__()
        self.downsample = downsample
        self.in_n = in_n
        self.mid_n = mid_n
        self.out_n = out_n
        self.kernel_size = kernel_size
        self.stride = 2 if self.downsample else 1

        self.shortcut = create_shortcut(in_n, out_n, stride=self.stride)
        self.ln2 = create_batch_norm(mid_n)  # LayerNorm1d(mid_n)
        self.ln3 = create_batch_norm(out_n)  # LayerNorm1d(out_n)

    def num_channels(self):
        return self.out_n + self.mid_n * 2

    def forward(self, x, all_W, all_b, from_idx):
        batch_size = x.shape[0]
        W1 = (
            all_W[:, from_idx : from_idx + self.mid_n]
            .view(batch_size, self.mid_n, self.kernel_size)
            .contiguous()
        )
        # Only use a subset of the full available layer.
        # W1 = W1[:, :, : self.in_n, :].contiguous()
        b1 = all_b[:, from_idx : from_idx + self.mid_n].view(batch_size, -1, 1)
        from_idx += self.mid_n
        W2 = (
            all_W[:, from_idx : from_idx + self.mid_n]
            .view(batch_size, self.mid_n, self.kernel_size)
            .contiguous()
        )
        # W2 = W2[:, :, : self.mid_n, :]
        from_idx += self.mid_n
        W3 = (
            all_W[:, from_idx : from_idx + self.out_n]
            .view(batch_size, self.out_n, self.kernel_size)
            .contiguous()
        )
        # W3 = W3[:, :, : self.mid_n, :]
        from_idx += self.out_n
        x = self._forward(x, W1, b1, W2, W3)
        return x, from_idx

    def _forward(self, x, W1, b1, W2, W3):
        shortcut = self.shortcut(x)
        x = F.relu(self.batch_unique_conv(x, W1) + b1)
        x = self.batch_unique_conv(x, W2, self.stride)
        x = F.relu(self.ln2(x))
        x = F.relu(self.ln3(self.batch_unique_conv(x, W3)))
        x = F.relu(x + shortcut)
        return x

    def batch_unique_conv(self, x, W, stride=1):
        batch_size = x.shape[0]
        W = W.view(-1, 1, self.kernel_size)
        num_out_channels = W.shape[0] // batch_size
        num_in_channels = x.shape[1]
        assert W.shape[0] % batch_size == 0
        # W = torch.reshape(W, shape=(-1, x.shape[1], self.kernel_size))
        x_flat = x.view(1, -1, x.shape[-1])
        x_flat = F.pad(x_flat, (self.kernel_size // 2, self.kernel_size // 2))
        x = F.conv1d(
            x_flat,
            W,
            bias=None,
            padding="valid",
            stride=stride,
            # Not sure about this!
            groups=batch_size * num_in_channels,
        )
        x = x.view(batch_size, num_out_channels, x.shape[-1])
        return x


class HyperResDecoder(nn.Module):
    def __init__(self, hyper_res_block, in_n, hidden1_n, hidden2_n):
        super().__init__()
        self.hyper_res_block = hyper_res_block
        self.fc1 = nn.Linear(in_n, hidden1_n)
        self.fc2 = nn.Linear(hidden1_n, hidden2_n)
        self.att1_shape = (
            self.hyper_res_block.mid_n,
            self.hyper_res_block.mid_warehouse_n,
        )
        self.att2_shape = (
            self.hyper_res_block.mid_n,
            self.hyper_res_block.mid_warehouse_n,
        )
        self.att3_shape = (
            self.hyper_res_block.out_n,
            self.hyper_res_block.out_warehouse_n,
        )
        self.fcA1 = nn.Linear(
            hidden2_n, self.att1_shape[0] * self.att1_shape[1]
        )
        self.fcA2 = nn.Linear(
            hidden2_n, self.att2_shape[0] * self.att2_shape[1]
        )
        self.fcA3 = nn.Linear(
            hidden2_n, self.att3_shape[0] * self.att3_shape[1]
        )
        self._reset_parameters()

    def _reset_parameters(self):
        nn.init.xavier_uniform_(self.fcA1.weight)
        nn.init.xavier_uniform_(self.fcA2.weight)
        nn.init.xavier_uniform_(self.fcA3.weight)
        self.fcA1.bias.data.fill_(0.0)
        self.fcA2.bias.data.fill_(0.0)
        self.fcA3.bias.data.fill_(0.0)

    def forward(self, x, z):
        batch_size = x.shape[0]
        h = torch.relu(self.fc1(z))
        h = torch.relu(self.fc2(h))
        key_scale = math.sqrt(h.shape[-1])
        att1 = F.softmax(
            self.fcA1(h).view(
                batch_size,
                self.hyper_res_block.mid_n,
                self.hyper_res_block.mid_warehouse_n,
            )
            / key_scale,
            dim=2,
        )
        att2 = F.softmax(
            self.fcA2(h).view(
                batch_size,
                self.hyper_res_block.mid_n,
                self.hyper_res_block.mid_warehouse_n,
            )
            / key_scale,
            dim=2,
        )
        att3 = F.softmax(
            self.fcA3(h).view(
                batch_size,
                self.hyper_res_block.out_n,
                self.hyper_res_block.out_warehouse_n,
            )
            / key_scale,
            dim=2,
        )
        x = self.hyper_res_block(x, att1, att2, att3)
        return x


class SEModule(nn.Module):
    def __init__(self, n_channels, reduction=1):
        super().__init__()
        n_mid = n_channels // reduction
        self.fc1 = nn.Conv1d(n_channels, n_mid, kernel_size=1)
        self.relu = nn.ReLU(inplace=True)
        self.fc2 = nn.Conv1d(n_mid, n_channels, kernel_size=1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        module_input = x
        x = x.mean((2,), keepdim=True)
        x = self.fc1(x)
        x = self.relu(x)
        x = self.fc2(x)
        x = self.sigmoid(x)
        return module_input * x


class HyperResBlock1d(nn.Module):
    def __init__(
        self,
        in_n,
        mid_n,
        out_n,
        warehouse_factor,
        kernel_size=3,
        downsample=False,
        squeeze=True,
    ):
        super().__init__()
        self.downsample = downsample
        self.in_n = in_n
        self.mid_n = mid_n
        self.out_n = out_n
        self.mid_warehouse_n = mid_n * warehouse_factor
        self.out_warehouse_n = out_n * warehouse_factor
        self.kernel_size = kernel_size
        stride = 2 if self.downsample else 1

        self.num_W1_weights = in_n * kernel_size
        self.num_W2_weights = mid_n * kernel_size
        self.num_W3_weights = mid_n * kernel_size

        self.shortcut = self.create_shortcut(in_n, out_n, stride=stride)
        self.W1_weights = nn.Linear(self.mid_warehouse_n, self.num_W1_weights)
        self.W2_weights = nn.Linear(self.mid_warehouse_n, self.num_W2_weights)
        self.W3_weights = nn.Linear(self.out_warehouse_n, self.num_W3_weights)

        # Bias is a way to inject channel information via attention, so
        # leave at least one layer using bias rather than batch norm.
        # Batch norm
        # self.bn1 = create_batch_norm(mid_n)
        self.bn2 = create_batch_norm(mid_n)
        self.bn3 = create_batch_norm(out_n)
        # Bias instead.
        self.b1_weights = nn.Linear(self.mid_warehouse_n, 1)
        # self.b2_weights = nn.Linear(mid_warehouse_n, 1)
        # self.b3_weights = nn.Linear(mid_warehouse_n, 1)
        self.se = SEModule(out_n)

    @staticmethod
    def create_shortcut(in_n, out_n, stride):
        # 1. Identity
        # In the simplest case, we can just return the input. For this to work,
        # both the channel count and channel dimensions of the input and ouput
        # must match. The output channel dimension is determined by the stride.
        channels_match = in_n == out_n
        downsample = stride > 1
        if channels_match and not downsample:
            return nn.Identity()

        skip_layers = []
        # 2. Downsample
        if downsample:
            # The following (1, 7) input:
            # |  1  |  2  |  3  |  4  |  5  |  6  |  7  |
            # when pooled with stride=kernel=2 becomes (1, 4):
            # |    1.5    |    3.5    |    5.5   | 7 |
            pool = nn.AvgPool1d(
                kernel_size=stride,
                stride=stride,
                count_include_pad=False,
                ceil_mode=True,
            )
            skip_layers.append(pool)
        # 3. 1x1 conv
        if not channels_match:
            # There isn't a consensus on whether:
            #   - to use batch norm or a bias or neither.
            conv = nn.Conv1d(in_n, out_n, kernel_size=1, bias=False)
            skip_layers.append(conv)
        res = nn.Sequential(*skip_layers)
        return res

    def forward(self, x, att1, att2, att3):
        batch_size = x.shape[0]
        shortcut = self.shortcut(x)
        # Attention
        W1 = self.W1_weights(att1).view(
            batch_size * self.mid_n, self.in_n, self.kernel_size
        )
        W2 = self.W2_weights(att2).view(
            batch_size * self.mid_n, self.mid_n, self.kernel_size
        )
        W3 = self.W3_weights(att3).view(
            batch_size * self.out_n, self.mid_n, self.kernel_size
        )
        b1 = self.b1_weights(att1).view(batch_size, self.mid_n)

        # b2 = self.b2_weights(att2).view(batch_size, self.mid_n)
        # b3 = self.b3_weights(att3).view(batch_size, self.out_n)
        # Conv
        # Use bias?
        x = F.relu(self.att_conv(x, W1, b1))
        x = F.relu(self.bn2(self.att_conv(x, W2)))
        x = F.relu(self.bn3(self.att_conv(x, W3)))
        x = self.se(x) + shortcut
        x = F.relu(x)
        return x

    def att_conv(self, x, W, b=None):
        batch_size = x.shape[0]
        x_flat = x.view(1, -1, x.shape[-1])
        x = F.conv1d(x_flat, W, bias=None, padding="same", groups=batch_size)
        num_out_channels = W.shape[0] // batch_size
        x = x.view(batch_size, num_out_channels, x.shape[-1])
        if b is not None:
            x = x + b.unsqueeze(-1)
        return x

        # x = F.relu(self.bn1(self.conv1(x)))
        # x = F.relu(self.bn2(self.conv2(x)))
        # x = self.bn3(self.conv3(x))
        # x = F.relu(x + shortcut)


class Attention(nn.Module):
    """
    Partly copied from:
        https://github.com/lucidrains/vit-pytorch/blob/main/vit_pytorch/simple_vit.py
    """

    def __init__(self, embed_dim, num_heads, head_dim):
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        inner_dim = self.num_heads * head_dim
        self.norm = nn.LayerNorm(self.embed_dim)
        self.to_qkv = nn.Linear(self.embed_dim, 3 * inner_dim, bias=False)
        self.to_out = nn.Linear(inner_dim, self.embed_dim, bias=False)

    def forward(self, x):
        x = self.norm(x)
        q, k, v = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(
            lambda t: einops.rearrange(
                t, "b n (h d) -> b h n d", h=self.num_heads
            ),
            [q, k, v],
        )
        attn = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.shape[-1]))
        attn = F.softmax(attn, dim=-1)
        y = attn @ v
        y = einops.rearrange(y, "b h n d -> b n (h d)")
        y = self.to_out(y)
        return y


class Transformer(nn.Module):
    def __init__(self, embed_dim, num_layers, num_heads, head_dim, mlp_dim):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(num_layers):
            self.layers.append(
                nn.ModuleDict(
                    {
                        "attn": Attention(embed_dim, num_heads, head_dim),
                        "ff": nn.Sequential(
                            nn.LayerNorm(embed_dim),
                            nn.Linear(embed_dim, mlp_dim),
                            nn.GELU(),
                            nn.Linear(mlp_dim, embed_dim),
                        ),
                    }
                )
            )

    def forward(self, x):
        for layer in self.layers:
            x = layer["attn"](x) + x
            x = layer["ff"](x) + x
        return x



from collections import namedtuple
from collections import deque
import itertools
import concurrent.futures
import logging
import math
from typing import (
    Dict,
    List,
    Tuple,
    Union,
    Sequence,
    Optional,
    Set,
    Iterable,
    Deque,
)
import pathlib
import pickle
import numpy as np
import pandas as pd
import scipy
import scipy.signal

import h5py


ELECTRODE_FREQ = 17852.767845719834  # Hz
NUM_STIMULUS_LEDS = 4
STIMULUS_PATTERN_FILENAME = "stimulus_pattern.h5"
SPIKE_RESPONSE_FILENAME = "spike_response.pickle"
RECORDED_STIMULUS_FILENAME = "recorded_stimulus.pickle"


Stimulus = namedtuple(
    "Stimulus", ["name", "wavelength", "channel", "display_hex", "import_name"]
)
stimuli = [
    Stimulus("red", 680, 0, "#FE7C7C", "/Red_Noise"),
    Stimulus("green", 530, 1, "#8AFE7C", "/Green_Noise"),
    Stimulus("blue", 470, 2, "#7CFCFE", "/Blue_Noise"),
    Stimulus("uv", 400, 3, "#7C86FE", "/UV_Noise"),
]
_stimulus_map = {s.name: s for s in stimuli}


def stimulus_by_name(name: str) -> Stimulus:
    """Return the stimulus info for the stimulus with the given name."""
    return _stimulus_map[name]


class CompressedSpikeRecording:
    """
    Data for a single recording, with spikes and stimulus stored as events.

    Storage as events means that a 10 bin long recording with a single spike
    at bin 5 would be have spike data = [5], as opposed to
    [0,0,0,0,1,0,0,0,0,0]. The recordings are stored in this format anyway,
    so this class is mostly just a wrapper around some Pandas dataframes. The
    SpikeRecording class is the decompressed version of this class.

    Some benefits of this class:

        - We are not sure what changes to the underlying dataframe we will be
          making in future, so this class provides a layer of abstraction.
        - The recording data is split into a triplet: (stimulus pattern,
          stimulus recording, spike recording) which typically needs to be
          moved around together. So combining them in a single place makes
          sense.
        - It's easy to make mistakes querying dataframes with the Pandas API.
          The queries can be done once here.


    Future work:

        - Currently, it's assumed that the stimulus pattern is full-field.
          Switching to a 2D stimulus might require some changes. At least we
          will need to decide the X-Y coordinate order and (0,0) position.
    """

    def __init__(
        self,
        name: str,
        stimulus_pattern: np.ndarray,
        stimulus_events: np.ndarray,
        spike_events: List[np.ndarray],
        cluster_ids: List[int],
        sensor_sample_rate: float,
        num_sensor_samples: int,
    ):
        if len(spike_events) != len(cluster_ids):
            raise ValueError(
                f"Mismatch between number of cluster-spikes "
                f"({len(spike_events)}) and number of cluster ids "
                f"({len(cluster_ids)})."
            )
        self.name = name
        self.stimulus_pattern = stimulus_pattern
        self.cluster_ids = cluster_ids
        self.stimulus_events = stimulus_events
        self.spike_events = spike_events
        self.sensor_sample_rate = sensor_sample_rate
        self.num_sensor_samples = num_sensor_samples

    def __str__(self):
        res = (
            f"Recording: {self.name}, "
            f"sensor sample rate: {self.sensor_sample_rate} Hz, "
            f"num samples: {self.num_sensor_samples}, "
            f"duration: {self.duration():.1f} seconds, "
            f"stimulus pattern shape: {self.stimulus_pattern.shape},"
            f"num clusters: {len(self.cluster_ids)}."
        )
        return res

    def duration(self):
        """Duration of the recording in seconds."""
        return self.num_sensor_samples / self.sensor_sample_rate

    def clusters(self, cluster_ids: Set[int]):
        """Returns a new recording with only the specified clusters."""
        if not cluster_ids.issubset(self.cluster_ids):
            raise ValueError(
                f"Cluster ids ({cluster_ids}) are not a subset of "
                f"the cluster ids in this recording ({self.cluster_ids})."
            )
        spike_events = [
            self.spike_events[self.cluster_ids.index(i)] for i in cluster_ids
        ]
        return CompressedSpikeRecording(
            self.name,
            self.stimulus_pattern,
            self.stimulus_events,
            spike_events,
            list(cluster_ids),
            self.sensor_sample_rate,
            self.num_sensor_samples,
        )


class SpikeRecording:
    """
    Data for a recording, with a spike and stimulus value for each time bin.

    The class allows indexing by time and cluster, so an object may represent
    a part of a recording, and isn't limited to be a whole recording.

    A 10 bin long recording with a single spike at bin 5 would be have spike
    data = [0,0,0,0,1,0,0,0,0,0], as opposed to [5]. The stimulus is stored
    similarly.

    This class is mostly just a wrapper around the already decompressed data.
    It's useful for the same reasons as the CompressedSpikeRecording class.

    Future work:

        - I'm slowly adding functionality to make this class more array like.
          The idea here is that the spike data and stimulus data have a common
          time dimension, which it is useful index and slice over, but
          otherwise their dimensions don't match. xarray would be a suitable
          off-the-shelf solution for this. xarray is probably best, but while
          the required functionality of this class is still minimal, the
          current solution seems fine, and saves having to require knowledge of
          yet another indexing API in addition to the currently used numpy,
          Pandas and PyTorch.
    """

    def __init__(self, name, stimulus, spikes, cluster_ids, sample_rate):
        if len(stimulus) != len(spikes):
            raise ValueError(
                f"Length of stimulus ({len(stimulus)}) and length"
                f" of response ({len(spikes)}) do not match."
            )
        if spikes.shape[1] != len(cluster_ids):
            raise ValueError(
                f"Mismatch between number of cluster-spikes "
                f"({spikes.shape[1]}) and number of cluster ids "
                f"({len(cluster_ids)})."
            )
        self.name = name
        self.stimulus = stimulus
        self.spikes = spikes
        self.cluster_ids = cluster_ids
        self.sample_rate = sample_rate

    def __str__(self):
        res = (
            f"Recording: {self.name}, "
            f"sample rate: {self.sample_rate:.1f} Hz "
            f"({1000/self.sample_rate:.3f} ms per sample), "
            f"duration: {self.duration():.1f} seconds, "
            f"stimulus shape: {self.stimulus.shape}, "
            f"spikes shape: {self.spikes.shape}, "
            f"num clusters: {len(self.cluster_ids)}."
        )
        return res

    def __len__(self):
        assert len(self.stimulus) == len(self.spikes)
        return len(self.stimulus)

    def duration(self):
        return self.stimulus.shape[0] / self.sample_rate

    def __getitem__(self, time_key):
        """Return a new recording with only data for the given time bin."""
        return SpikeRecording(
            f"{self.name}-{str(time_key)}",
            self.stimulus[time_key],
            self.spikes[time_key],
            self.cluster_ids,
            self.sample_rate,
        )

    def clusters(self, cluster_ids: Set[int]):
        """Returns a new recording with only the specified clusters."""
        if not cluster_ids.issubset(self.cluster_ids):
            raise ValueError(
                f"Cluster ids ({cluster_ids}) are not a subset of "
                f"the cluster ids in this recording ({self.cluster_ids})."
            )
        cluster_indices = [self.cluster_ids.index(i) for i in cluster_ids]
        return SpikeRecording(
            self.name,
            self.stimulus,
            self.spikes[:, cluster_indices],
            cluster_ids,
            self.sample_rate,
        )

    def extend(self, recording):
        """Add the given recording to the end of this one.

        This was added in order to create test, train and validation sets that
        pick their data from multiple parts of the same recording.
        """
        if self.sample_rate != recording.sample_rate:
            raise ValueError(
                f"Sample rates do not match. Got ({self.sample_rate}) and "
                f"({recording.sample_rate})."
            )
        if self.stimulus.shape[1] != recording.stimulus.shape[1]:
            raise ValueError(
                f"The stimulus must have the same number of LEDs. Got "
                f"({self.stimulus.shape[1]}) and "
                f"({recording.stimulus.shape[1]})."
            )
        if not np.array_equal(self.cluster_ids, recording.cluster_ids):
            raise ValueError(
                f"Cluster ids do not match. Got ({self.cluster_ids}) and "
                f"({recording.cluster_ids})."
            )
        self.stimulus = np.concatenate((self.stimulus, recording.stimulus))
        self.spikes = np.concatenate((self.spikes, recording.spikes))
        return self

    def spike_snippets(self, total_len: int, post_spike_len: int):
        snippets_by_cluster = []
        # Can this be done in a single call?
        # Could make compress_spikes operate on 2d array, then sent all
        # spikes to spike_snippets then split with np.split().
        for c in range(self.spikes.shape[1]):
            snippets_by_cluster.append(
                spike_snippets(
                    self.stimulus,
                    compress_spikes(self.spikes[:, c]),
                    total_len,
                    post_spike_len,
                )
            )
        return snippets_by_cluster


def split(recording: SpikeRecording, split_ratio: Sequence[int]):
    """Split a recording into multiple recordings.

    Args:
        split_ratio: a list of weightings that determines how much data to
            give each split. For example, you might use the triplet (3, 1, 1)
            to create a train-val-test split with 60% of the data for training,
            20% for validation, and 20% for testing.
    """
    if len(split_ratio) < 2:
        raise ValueError("Can't split a recording into fewer than 2 parts.")
    if not all([r > 0 for r in split_ratio]):
        raise ValueError(f"Split ratios must be positive. Got ({split_ratio}).")
    divisions = sum(split_ratio)
    num_per_division, remainder = divmod(len(recording), divisions)
    splits = []
    slice_start = slice_end = 0
    for i in range(len(split_ratio)):
        if i == 0:
            # Give all of the remainder to the first split.
            slice_end = split_ratio[i] * num_per_division + remainder
        else:
            slice_end += split_ratio[i] * num_per_division
        splits.append(recording[slice_start:slice_end])
        slice_start = slice_end
    total_len = sum([len(s) for s in splits])
    assert total_len == len(
        recording
    ), f"Split lengths do not match ({total_len}) vs ({len(recording)})."
    return splits


def mirror_split(recording: SpikeRecording, split_ratio: Sequence[int]):
    """Split data from "outside-in".

    This approach of splitting the data tries to address the issue that the
    response of the retina will change over time. Training the model with the
    earlier responses and validating or testing with the later responses will
    likely lead to reduced accuracy that is not the fault of the model.

    One way to ammeliorate this issue is to give each split a piece of the
    earlier data and a piece of the later data. There is still some flexibility
    in how to make this choice. If there are three splits (train, val, test),
    then they will be made as follows:

        +-------+------+----+----+------+-------+
        | train | val  |  test   | val  | train |
        +-------+------+----+----+------+-------+

    The benefit of this approach is that each split is not exposed soley to
    one end of the data. Additionally, the test data, which will often appear
    in reports will still exist as a continuous block of data.

    Args:
        split_ratio: a list of weightings that determines how much data to
            give each split. For example, you might use the triplet (3, 1, 1)
            to create a train-val-test split with 60% of the data for training,
            20% for validation, and 20% for testing.
    """
    if not all([r > 0 for r in split_ratio]):
        raise ValueError(f"Split ratios must be positive. Got ({split_ratio}).")
    recording_half1 = recording[: len(recording) // 2]
    recording_half2 = recording[len(recording) // 2 :]
    assert len(recording_half1) + len(recording_half2) == len(recording)
    splits_half1 = split(recording_half1, split_ratio)
    splits_half2 = split(recording_half2, tuple(reversed(split_ratio)))
    splits = [
        s1.extend(s2) for (s1, s2) in zip(splits_half1, reversed(splits_half2))
    ]
    total_len = sum([len(s) for s in splits])
    assert total_len == len(
        recording
    ), f"Split lengths do not match ({total_len}) vs ({len(recording)})."
    return splits


def load_stimulus_pattern(file_path: Union[pathlib.Path, str]) -> np.ndarray:
    """
    Loads the stimulus data from the HDF5 file as a Pandas DataFrame.

    Dataframe structure
    -------------------
        - integer index, representing stimulus frames.
        - columns are ['red', 'green', 'blue', 'uv']
        - values are 0 or 1 representing whether the corresponding LED is ON
          or OFF at the given stimulus frame.
    """
    file_path = pathlib.Path(file_path)
    with h5py.File(file_path, "r") as f:
        # The data has shape: [4, 24000, 10374]. This corresponds to 4 lights,
        # on-off pattern for 20min at 20 Hz (24000 periods), and 10374 boxes
        # arranged over 2D screen space. In the full-field experiment, only a
        # single box was used, and hence the [:,0] access pattern.
        colour_noise = np.array(
            [
                f[stimuli[0].import_name][:, 0],
                f[stimuli[1].import_name][:, 0],
                f[stimuli[2].import_name][:, 0],
                f[stimuli[3].import_name][:, 0],
            ]
        ).transpose()
    stimulus_switch_freq = 20
    stimulus_loop_mins = 20
    stimulus_loop_secs = stimulus_loop_mins * 60
    assert colour_noise.shape[0] == stimulus_loop_secs * stimulus_switch_freq
    # Optional, if bottleneck encountered
    # ===================================
    # We are doing a lot of slicing in the time-step dimension, so keep it
    # as the row dimension. Also insure the array is contiguous for further
    # speed-ups.
    # res = np.ascontiguousarray(colour_noise)
    return colour_noise


def load_response(
    file_path: Union[pathlib.Path, str], keep_kernels=False
) -> pd.DataFrame:
    """Loads the spike data from a Pickle file as a Pandas DataFrame.

    The input path should point to standard pickle file (zipped or not).

    Dataframe structure
    -------------------
    The dataframe uses a multiindex: ['Cell index', 'Stimulus ID', 'Recording'].
        - The cell index refers to the id assigned by the spike sorter.
        - TODO: The stimulus ID is what?
        - The 'Recording' index is a human readable string to identify a
            recording session.

    The Dataframe has 2 columns: ['Kernel', 'Spikes'].
        - The each row contains a (2000,4) shaped Numpy array holding the
          precomputed response kernel for the matching cell-stimulus-recording.
          The kernel represents 2000 miliseconds, with the spike located at
          1000 ms.
        - The 'Spikes' column contains a (17993,) shaped *masked* Numpy array,
          holding a variable number of integers. Each integer reprents the
          time (in sensor readings) at which the spike occurred.
    """
    res = pd.read_pickle(file_path)
    if not keep_kernels:
        res.drop("Kernel", axis=1)
    return res


def load_recorded_stimulus(file_path: Union[pathlib.Path, str]) -> pd.DataFrame:
    """Load the spike data from a Pickle file as a Pandas DataFrame."""
    res = pd.read_pickle(file_path, compression="infer")
    return res


def stim_and_spike_rows(
    rec_name: str, stimulus_df: pd.DataFrame, response_df: pd.DataFrame
) -> Tuple[pd.Series, pd.Series]:
    """
    Return the stimulus row and spike rows corresponding to a single recording.
    """
    SUPPORTED_STIMULUS = 1
    stim_row = stimulus_df.xs(
        (SUPPORTED_STIMULUS, rec_name), level=("Stimulus_index", "Recording")
    )  # .reset_index("Stimulus_index", drop=True)
    if len(stim_row) != 1:
        raise ValueError(
            "There should be only one stimulus recording. "
            f"Got ({len(stim_row)}). For recording name: ({rec_name})"
        )
    stim_row = stim_row.iloc[0]
    response_rows = response_df.xs(
        (SUPPORTED_STIMULUS, rec_name), level=("Stimulus ID", "Recording")
    )  # .reset_index("Stimulus ID", drop=True)
    return stim_row, response_rows


def _assert_file_exists(p: pathlib.Path):
    if not p.exists():
        raise ValueError(f"File does not exist: ({p}).")
    if not p.is_file():
        raise ValueError(f"Path is not a file: ({p}).")


def _load_data_files(data_dir: pathlib.Path):
    stimulus_pattern_path = data_dir / STIMULUS_PATTERN_FILENAME
    stimulus_recording_path = data_dir / RECORDED_STIMULUS_FILENAME
    response_recording_path = data_dir / SPIKE_RESPONSE_FILENAME
    _assert_file_exists(stimulus_pattern_path)
    _assert_file_exists(stimulus_recording_path)
    _assert_file_exists(response_recording_path)
    stim_pattern = load_stimulus_pattern(stimulus_pattern_path)
    stim_recordings = load_recorded_stimulus(stimulus_recording_path)
    response_recordings = load_response(response_recording_path)
    return stim_pattern, stim_recordings, response_recordings


def load_3brain_recordings(
    data_dir: Union[str, pathlib.Path],
    include: Optional[Iterable[str]] = None,
    num_workers: int = 4,
) -> List[CompressedSpikeRecording]:
    """
    Creates a CompressedSpikeRecording for each recording in the 3Brain data.

    Args:
        include: a list of recording names to include.
        num_workers: how many threads can be used to load the data.
    """
    # Load the data. Do it once here.
    data_dir = pathlib.Path(data_dir)
    (
        stimulus_pattern,
        stimulus_recordings,
        response_recordings,
    ) = _load_data_files(data_dir)
    rec_names = recording_names(response_recordings)
    # Collect the recordings to load.
    to_load = []
    for rec_name in rec_names:
        do_load = include is None or rec_name in include
        if do_load:
            to_load.append(rec_name)
    done = []

    def _load(rec_name):
        rec_obj = _single_3brain_recording(
            rec_name, stimulus_pattern, stimulus_recordings, response_recordings
        )
        return rec_obj

    with concurrent.futures.ThreadPoolExecutor(
        max_workers=num_workers
    ) as executor:
        future_to_name = {
            executor.submit(_load, rec_name): rec_name for rec_name in to_load
        }
        for future in concurrent.futures.as_completed(future_to_name):
            rec_name = future_to_name[future]
            try:
                data = future.result()
            except Exception as exc:
                print(f"Loading {rec_name} generated an exception: {exc}.")
            else:
                done.append(data)
    return done


def single_3brain_recording(
    rec_name: str,
    data_dir: Union[pathlib.Path, str],
    include_clusters: Optional[Set[int]] = None,
) -> CompressedSpikeRecording:
    data_dir = pathlib.Path(data_dir)
    (
        stimulus_pattern,
        stimulus_recordings,
        response_recordings,
    ) = _load_data_files(data_dir)
    return _single_3brain_recording(
        rec_name,
        stimulus_pattern,
        stimulus_recordings,
        response_recordings,
        include_clusters,
    )


def _single_3brain_recording(
    rec_name: str,
    stimulus_pattern: np.ndarray,
    stimulus_recordings: pd.DataFrame,
    response_recordings: pd.DataFrame,
    include_clusters: Optional[Set[int]] = None,
) -> CompressedSpikeRecording:
    stimulus_row, response_rows = stim_and_spike_rows(
        rec_name, stimulus_recordings, response_recordings
    )
    # Handle stimulus.
    if stimulus_row["Stimulus_name"] != "FF_Noise":
        raise ValueError(
            f'Only stimulus type "FF_NOISE" is currently '
            f'supported. Got ({stimulus_row["Stimulus_name"]})'
        )
    if include_clusters and not include_clusters.issubset(
        set(response_rows.index)
    ):
        raise ValueError(
            f"Cluster ids ({include_clusters}) are not a subset of "
            f"the cluster ids in this recording {tuple(response_rows.index)}."
        )
    # TODO: is 'End_Fr' inclusive? If so, add 1 below. Assuming yes for now.
    stimulus_events = stimulus_row["Trigger_Fr_relative"].astype(int)
    num_samples = stimulus_events[-1] + 1
    # Remove the last event, as it's the end.
    stimulus_events = stimulus_events[:-1]
    sensor_sample_rate = stimulus_row["Sampling_Freq"]

    # Handle spikes.
    spikes_per_cluster = []
    cluster_ids = []
    for cluster_id, cluster_row in response_rows.iterrows():
        if include_clusters is not None and cluster_id not in include_clusters:
            continue
        spikes = cluster_row["Spikes"].compressed().astype(int)
        assert not np.ma.isMaskedArray(spikes), "Don't forget to decompress!"
        spikes_per_cluster.append(spikes),
        cluster_ids.append(cluster_id)

    # Create return object.
    res = CompressedSpikeRecording(
        rec_name,
        stimulus_pattern,
        stimulus_events,
        spikes_per_cluster,
        cluster_ids,
        sensor_sample_rate,
        num_samples,
    )
    return res


def decompress_recordings(
    recordings: Iterable[CompressedSpikeRecording],
    downsample: int = 1,
    num_workers=10,
) -> Deque[SpikeRecording]:
    """
    Decompress multiple recordings.
    """

    def _decompress(rec):
        return decompress_recording(rec, downsample)

    with concurrent.futures.ThreadPoolExecutor(
        max_workers=num_workers
    ) as executor:
        res = deque(executor.map(_decompress, recordings))
        executor.shutdown(wait=True)
    return res


def decompress_recording(
    recording: CompressedSpikeRecording, downsample: int
) -> SpikeRecording:
    """
    Decompress a compressed recording.

    The result holds numpy arrays where the first dimension is time.

    Without downsampling, a single recording take up more than 1 gigabyte of
    memory. It's quite convenient to set the  downsample to 18, as this will
    cause the resulting sample rate to be 991.8 Hz, which is the closest you
    can get to 1 kHz, given the 3Brain electrode's original frequency.
    """
    sample_rate = recording.sensor_sample_rate / downsample
    stimulus = decompress_stimulus(
        recording.stimulus_pattern,
        recording.stimulus_events,
        recording.num_sensor_samples,
        downsample,
    )
    spikes = np.stack(
        [
            decompress_spikes(s, recording.num_sensor_samples, downsample)
            for s in recording.spike_events
        ],
        axis=1,
    )
    res = SpikeRecording(
        recording.name, stimulus, spikes, recording.cluster_ids, sample_rate
    )
    return res


def decompress_stimulus(
    stimulus_pattern: np.ndarray,
    trigger_events: np.ndarray,
    total_length: int,
    downsample: int,
) -> np.ndarray:
    if trigger_events[0] != 0:
        raise ValueError(
            "The trigger events are expected to start at zero, "
            f"but the first trigger was at ({trigger_events[0]})."
        )
    if len(trigger_events) > len(stimulus_pattern):
        raise ValueError(
            "Recorded stimulus is longer than the stimulus "
            f"pattern. ({len(trigger_events)} > {len(stimulus_pattern)})"
        )
    # TODO: check assumption! Assuming that the stimulus does not continue
    # after the last trigger event. This makes the last trigger special in that
    # it doesn't mark the start of a new stimulus output.
    # TODO: Marvin says that the last trigger event isn't the end of the
    # recording, but the last stimulus event before the end of the recording.
    # This introduces issues in that it's not clear how long this final
    # stimulus event is. I'm keeping the functionality as it is now, as I think
    # a better solution is to encode the recording end in the last stimulus
    # event, which would make the existing code below work fine.
    logging.info(
        f"Starting: decompressing stimulus. Resulting shape ({total_length})."
    )
    # If this becomes a bottleneck, there are some tricks to reach for:
    # https://stackoverflow.com/questions/60049171/fill-values-in-numpy-array-that-are-between-a-certain-value
    num_channels = stimulus_pattern.shape[1]
    res = np.empty(shape=(total_length, num_channels))
    # Pair each trigger with the next trigger, and do this for all but the last.
    slices = np.stack((trigger_events[:-1], trigger_events[1:]), axis=1)
    for idx, s in enumerate(slices):
        res[np.arange(*s)] = stimulus_pattern[idx]
    last_trigger = trigger_events[-1]
    res[last_trigger:] = stimulus_pattern[len(slices)]
    logging.info(
        f"Finished: decompressing stimulus. "
        f"The last trigger was at ({trigger_events[-1]}) making its "
        f"duration ({res.shape[0]} - {trigger_events[-1]} = "
        f"{res.shape[0] - trigger_events[-1]}) samples."
    )
    res = downsample_stimulus(res, downsample)
    return res


def rebin_spikes(spike_idxs: np.ndarray, downsample_factor: int) -> np.ndarray:
    """Calculate the spike indices after downsampling.

    Args:
        spike_idxs: The spike indices before downsampling.

    This method is very simple: just floor divide the indicies. It's good to
    have a dedicated function just so we are clear what the behaviour is,
    and to insure that we are doing it consistently.
    """
    res = np.floor_divide(spike_idxs, downsample_factor)
    return res


def decompress_spikes(
    spikes: Union[np.ndarray, np.ma.MaskedArray],
    num_sensor_samples: int,
    downsample_factor: int = 1,
) -> np.ndarray:
    """
    Fills an integer array counting the number of spikes that occurred.

    If downsample_factor is 1 (no downampling), then the output array will
    be an array of 0s and 1s, where 1 indicates that a spike occurred in that
    time bin.

    Setting downsample_factor to an integer greater than 1 will result in
    the spikes being counted in larger bin sizes that the original sensor
    sample period. So we are not talking about signal downsampling, Nyquist
    rates etc., rather we are talking about histogram binning where the bin
    size is scaled by downsample_factor. This behaviour is similar to Pandas's
    resample().sum() pattern.

    Binning behaviour
    -----------------
    As only integer values are accepted for downsample_factor, the binning is
    achieved by floor division of the original spike index. Examples:
        1. Input: [0, 0, 0, 1, 1], downsample_factor=2, output: [0, 1, 1]
        1. Input: [0, 0, 0, 1, 1, 1], downsample_factor=2, output: [0, 1, 2]
    """
    if np.ma.isMaskedArray(spikes):
        spikes = spikes.compressed()
    downsampled_spikes = rebin_spikes(spikes, downsample_factor)
    res = np.zeros(
        shape=[
            math.ceil(num_sensor_samples / downsample_factor),
        ],
        dtype=int,
    )
    np.add.at(res, downsampled_spikes, 1)
    return res


def factors_sorted_by_count(
    n, limit: Optional[int] = None
) -> Tuple[Tuple[int, ...]]:
    """
    Calulates factor decomposition with sort and limit.

    This method is used to choose downsampling factors when a single factor
    is too large. The decompositions are sorted by the number of factors in a
    decomposition. With this in mind, when a factorization cannot be found
    that has all factors

    Args:
        n: The number to decompose.
        limit: The maximum number of factors allowed in a single decomposition.
            This is an inclusive limit. If there are no decompositions t
    """

    def _factors(n):
        # Use a set to avoid duplicate factors.
        res = {(n,)}
        f1 = n // 2
        while f1 > 1:
            f2, mod = divmod(n, f1)
            if not mod:
                res.add(tuple(sorted((f1, f2))))
                for (a, b) in itertools.product(_factors(f1), _factors(f2)):
                    sub_factors = tuple(sorted(a + b))
                    res.add(sub_factors)
            f1 -= 1
        return res

    factors = _factors(n)
    sorted_by_count = tuple(sorted(factors, key=lambda x: len(x)))
    # If there is a limit set, remove any factor decompositions that contain a
    # factor larger than the limit.
    if limit:
        factors_filtered = tuple(f for f in factors if max(f) <= limit)
        if not factors_filtered:
            logging.info(
                f"No factor decomposition exists with factors under "
                f"{limit}. Returning the decomposition with the "
                f"most factors."
            )
            sorted_by_count = (sorted_by_count[-1],)
        else:
            sorted_by_count = factors_filtered
    return sorted_by_count


def downsample_stimulus(stimulus: np.ndarray, factor: int) -> np.ndarray:
    """
    Filter (low-pass) a stimulus and then decimate by a factor.

    This is needed to prevent aliasing.

    Resources on filtering
    ----------------------
    https://dsp.stackexchange.com/questions/45446/pythons-tt-resample-vs-tt-resample-poly-vs-tt-decimate
    https://dsp.stackexchange.com/questions/83696/downsample-a-signal-by-a-non-integer-factor
    https://dsp.stackexchange.com/questions/83889/decimate-a-signal-whose-values-are-calculated-not-stored?noredirect=1#comment176944_83889
    """
    if factor == 1:
        return stimulus
    time_axis = 0
    logging.info(
        f"Starting: downsampling by {factor}. Initial length "
        f"{stimulus.shape[time_axis]}."
    )
    # SciPy recommends to never exceed 13 on a single decimation call.
    # See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.decimate.html
    MAX_SINGLE_DECIMATION = 13
    sub_factors = factors_sorted_by_count(factor, limit=MAX_SINGLE_DECIMATION)[
        0
    ]
    for sf in sub_factors:
        logging.info(f"Starting: decimating by {sf}")
        stimulus = scipy.signal.decimate(
            stimulus, sf, ftype="fir", axis=time_axis
        )
        logging.info(f"Finished: decimating by {sf}")
    logging.info(
        f"Finished: downsampling. Resulting length "
        f"({stimulus.shape[time_axis]})."
    )
    return stimulus


def recording_names(response: pd.DataFrame) -> list:
    """Return the list of recording names."""
    rec_list = response.index.get_level_values("Recording").unique().tolist()
    return rec_list


def _list_to_index_map(l) -> Dict[str, int]:
    """
    Returns a map from list elements to their index in the list.
    """
    return {x: i for i, x in enumerate(l)}


def _cluster_ids(response: pd.DataFrame, rec_name: str) -> list:
    """Returns the list of cluster ids for a given recording.

    Stimulus ID = 7 is currently ignored (actually, anything other than 1).

    This function isn't used anymore, I don't think. Keeping it around for
    reference, for a little while.
    """
    # Note: I'm not sure if it's better to request by recording name or
    # the recording id.
    stimulus_id = 1  # not 7.
    cluster_ids = (
        response.xs(
            (stimulus_id, rec_name),
            level=("Stimulus ID", "Recording"),
            drop_level=True,
        )
        .index.get_level_values("Cell index")
        .unique()
        .tolist()
    )
    return cluster_ids


def spike_window(
    spike_idxs: Union[int, Sequence[int]], total_len: int, post_spike_len: int
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculate the window endpoints around a spike, in samples of the stimulus.
    """
    if total_len < post_spike_len + 1:
        raise ValueError(
            f"Total snippet length must be at least 1 greater than the "
            f"post-spike length + 1. Got total_len ({total_len}) "
            f"and post_spike_len ({post_spike_len})."
        )
    spikes = np.array(spike_idxs)
    # Calculate the snippet start and end.
    # The -1 appears as we include the spike sample in the snippet.
    win_start = (spikes + post_spike_len) - (total_len - 1)
    win_end = spikes + post_spike_len + 1
    return win_start, win_end


def spike_snippets(
    stimulus: np.ndarray,
    spike_idxs: np.ndarray,
    total_len: int,
    post_spike_len: int,
) -> np.ndarray:
    """
    Return a subset of the stimulus around the spike points.

    Args:
        stimulus: The decompressed stimulus.
        spikes: Spike counts, re-binned to the stimulus sample rate.
        total_len: The length of the snippet in stimulus frames.
        post_spike_len: The number of frames to pad after the spike.
    Returns:
        A Numpy array of shape (spikes.shape[0], total_len, NUM_STIMULUS_LEDS).

    Note 1: The total length describes the snippet length inclusive of the post
        spike padding.
    Note 2: If a spike bin has 2 spikes, then the snippet they share will be
        added to the output twice.
    Note 3: If a spike happens early enough that the snippet would start before
        the stimulus, then the snippet will be padded with zeros. This applies
        to the end of the stimulus as well.

    Single spike example
    ====================

        frame #:  |   0   |   1   |   2   |   3   |   4   |   5   |   6   |   7   |
        ===========================================================================
        stimulus: |   0   |   1   |   1   |   0   |   0   |   1   |   0   |   0   |
                  |   1   |   1   |   0   |   0   |   1   |   0   |   0   |   0   |
                  |   1   |   1   |   1   |   0   |   0   |   0   |   0   |   1   |
                  |   0   |   0   |   0   |   0   |   1   |   0   |   1   |   1   |
        ===========================================================================
        spikes:   |   0   |   0   |   0   |   0   |   1   |   0   |   0   |   0   |

    The slice with parameters:
        - total length = 5
        - post spike length = 1

    Would be:

                  |   1   |   1   |   0   |   0   |   1   |
                  |   1   |   0   |   0   |   1   |   0   |
                  |   1   |   1   |   0   |   0   |   0   |
                  |   0   |   0   |   0   |   1   |   0   |
    """
    if np.ma.isMaskedArray(spike_idxs):
        raise ValueError(
            "spikes must be a standard numpy array, not a masked array."
        )
    # 1. Get the spike windows.
    win_start, _ = spike_window(spike_idxs, total_len, post_spike_len)
    # 2. Pad the stimulus in case windows go out of range.
    if np.any(win_start < 0) or np.any(
        win_start >= (stimulus.shape[0] - total_len)
    ):
        stimulus = np.pad(
            stimulus, ((total_len, total_len), (0, 0)), "constant"
        )
        # 3. Offset the windows, which is needed due to the padding.
        win_start += total_len
    # 4. Extract the slice.
    # The padded_stimulus is indexed by a list arrays of the form:
    #    (win_start[0], win_start[0]+1, win_start[0]+2, ..., win_start[0]+total_len)
    #    (win_start[1], win_start[1]+1, win_start[1]+2, ..., win_start[1]+total_len)
    #    ...
    #    (win_start[num_spikes-1], win_start[num_spikes-1]+1, win_start[num_spikes-1]+2, ..., win_start[num_spikes-1]+total_len)
    snippets = stimulus[np.asarray(win_start)[:, None] + np.arange(total_len)]
    return snippets


def compress_spikes(spikes: np.ndarray) -> np.ndarray:
    """Converts a 1D spike array to an index of spike array.

    Example:
        [0, 0, 1, 0, 0, 2, 3] => [2, 5, 5, 6, 6, 6]
    """
    nonzero_idxs = np.squeeze(np.nonzero(spikes))
    # Alternative:
    #   return np.repeat(np.arange(len(spikes)), spikes)
    return np.repeat(nonzero_idxs, spikes[nonzero_idxs])


def labeled_spike_snippets(
    rec: CompressedSpikeRecording,
    snippet_len: int,
    snippet_pad: int,
    downsample: int = 1,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Caluclate spike snippets for a recording, paired with the cluster ids.

    Args:
        snippet_len: the length of the snippet.
        snippet_pad: the number of timesteps to include after the spike.

    Returns: two np.ndarrays tuple. The first element contains
    the spike snippets and the second element contains ids of the clusters.
    """
    stimulus = decompress_stimulus(
        rec.stimulus_pattern,
        rec.stimulus_events,
        rec.num_sensor_samples,
        downsample,
    )

    # Gather the snippets and align them with a matching array of cluster ids.
    snippets = []
    cluster_ids = []
    for idx, cluster_id in enumerate(rec.cluster_ids):
        spike_idxs = rebin_spikes(rec.spike_events[idx], downsample)
        snips = spike_snippets(
            stimulus,
            spike_idxs,
            snippet_len,
            snippet_pad,
        )
        cluster_ids.extend(
            [
                cluster_id,
            ]
            * len(snips)
        )
        snippets.extend(snips)
    snippets = np.stack(snippets)
    cluster_ids = np.array(cluster_ids)
    return snippets, cluster_ids



import logging
import pytest
import retinapy.spikeprediction as sp
import retinapy.mea as mea
import itertools
import pathlib

DATA_DIR = pathlib.Path("./data/ff_noise_recordings")


# TODO: move this to common test utils.
@pytest.fixture
def recs():
    res = mea.load_3brain_recordings(
        DATA_DIR,
        include=["Chicken_04_08_21_Phase_01", "Chicken_04_08_21_Phase_02"],
    )
    # Filter out some clusters
    res[0] = res[0].clusters({40, 41})
    res[1] = res[1].clusters({46, 352, 703})
    return res


def test_trainable_factories(recs):
    """
    Tests multiple functions in one go (so as to speed up tests).

    Tests that:
        1. for each tranable group, a trainable is created without error for
        a small set of different configurations.
    """
    # Setup
    downsample_factors = [89, 178]
    input_lengths_ms = [992, 1586]
    output_lenghts_ms = [1, 50]
    configs = tuple(
        sp.Configuration(*tple)
        for tple in itertools.product(
            downsample_factors,
            input_lengths_ms,
            output_lenghts_ms,
        )
    )
    # Get default options from arg parser.
    _, parser = sp.arg_parsers()
    default_opts = parser.parse_args([])
    # For the models that only support a single cluster:
    single_cluster = [recs[0].clusters({40})]

    # Test
    for config in configs:
        assert (
            sp.LinearNonLinearTGroup.create_trainable(
                single_cluster, config, default_opts
            )
        ) is not None
        assert (
            sp.DistFieldCnnTGroup.create_trainable(
                single_cluster, config, default_opts
            )
        ) is not None
        assert (
            sp.MultiClusterDistFieldTGroup.create_trainable(
                recs, config, default_opts
            )
        ) is not None
        assert (
            sp.TransformerTGroup.create_trainable(recs, config, default_opts)
        ) is not None
        assert (
            sp.ClusteringTGroup.create_trainable(recs, config, default_opts)
        ) is not None



import torch
import torch.nn as nn
from torch.nn import functional as F
from typing import Union
import pathlib
import einops
import einops.layers.torch
import math

import retinapy
import retinapy.mea as mea
import retinapy.nn

import logging

_logger = logging.getLogger(__name__)


def load_model(
    model, checkpoint_path: Union[str, pathlib.Path], map_location=None
):
    checkpoint_path = pathlib.Path(checkpoint_path)
    if not checkpoint_path.exists():
        raise FileNotFoundError(
            f"Checkpoint file/folder ({checkpoint_path}) not found."
        )
    if checkpoint_path.is_dir():
        checkpoint_path = list(checkpoint_path.glob("*.pth"))[-1]

    _logger.info(f"Loading model from ({checkpoint_path}).")
    checkpoint_state = torch.load(checkpoint_path, map_location)
    model_state = checkpoint_state["model"]
    model.load_state_dict(model_state)

def load_model_and_optimizer(
    model, optimizer, checkpoint_path: Union[str, pathlib.Path]):
    checkpoint_path = pathlib.Path(checkpoint_path)
    if not checkpoint_path.exists():
        raise FileNotFoundError(
            f"Checkpoint file/folder ({checkpoint_path}) not found."
        )
    if checkpoint_path.is_dir():
        checkpoint_path = list(checkpoint_path.glob("*.pth"))[-1]

    _logger.info(f"Loading model from ({checkpoint_path}).")
    checkpoint_state = torch.load(checkpoint_path)
    model_state = checkpoint_state["model"]
    model.load_state_dict(model_state)
    optimizer.load_state_dict(checkpoint_state["optimizer"])


def save_model(model, path: pathlib.Path, optimizer=None):
    _logger.info(f"Saving model to ({path})")
    if not path.parent.exists():
        path.parent.mkdir(parents=True)
    state = {
        "model": model.state_dict(),
    }
    if optimizer:
        state.update({"optimizer": optimizer.state_dict()})
    torch.save(state, path)


def dist_loss(prediction, target):
    """
    MSE loss reduced along time dimension but not batch dimension.

    It is expected that this loss is just one part of a loss term, so reducing
    over the batch dimension is not done. And it's not even given as a option
    as it's too easy to mistake "mean" reduction to be over a single dimension,
    when it is a full reduction over all dimensions. This is a tricky aspect of
    PyTorch's MSE loss.
    """
    # Scale to get roughly in the ballpark of 0.1 to 10.
    DIST_LOSS_SCALE = 3
    loss = DIST_LOSS_SCALE * F.mse_loss(prediction, target, reduction="none")
    assert len(prediction.shape) == 2, "Batch and time dim expected."
    time_ave = torch.mean(loss, dim=1)
    batch_sum = torch.sum(time_ave)
    return batch_sum


class DistLossLinear(nn.Module):
    def __init__(self, alpha=0.1, beta=0.01):
        super().__init__()
        self.alpha = alpha
        self.beta = beta

    def forward(self, prediction, target):
        # In some sense, the target is never 0. A best estimate is that it's
        # closer to this time value than the previous or next. In expectation,
        # the distance will be 0.25 away from this time point, with a
        # uniform prior.
        error = F.l1_loss(prediction, target)
        scaling = self.alpha + 1 / (self.beta + target)
        loss = torch.mean(error * scaling)
        return loss


class IntervalWeightedDistLoss(nn.Module):
    def __init__(self, alpha=0.1, beta=0.05):
        super(IntervalWeightedDistLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta

    def forward(self, prediction, target, interval):
        error = F.l1_loss(prediction, target)
        scaling = self.alpha + 1 / (self.beta + interval)
        loss = torch.mean(error * scaling)
        return loss


class FcModel(nn.Module):
    def __init__(self, in_n, out_n, clamp_max):
        super(FcModel, self).__init__()
        self.clamp_max = clamp_max
        self.fc_body = retinapy.nn.FcBlock(
            hidden_ch=1024 * 2,
            num_hidden_layers=5,
            in_features=in_n,
            out_features=out_n,
            outermost_linear=True,
        )
        self.act = torch.nn.Softplus()

    def forward(self, x):
        x = self.act(self.fc_body(x))
        x = torch.clamp(x, min=None, max=self.clamp_max)
        return x


class MiniVAE(nn.Module):
    def __init__(self, num_clusters, z_n=2):
        super(MiniVAE, self).__init__()
        self.embed_mu = nn.Embedding(num_clusters, z_n)
        self.embed_logvar = nn.Embedding(num_clusters, z_n)

    def encode(self, x):
        x = x.long()
        mu = self.embed_mu(x)
        logvar = self.embed_logvar(x)
        return mu, logvar

    def sampling(self, mu, logvar):
        std = logvar.mul(0.5).exp_()
        eps = std.new_empty(std.size()).normal_()
        return eps.mul_(std).add_(mu)

    def forward(self, x):
        z_mu, z_logvar = self.encode(x)
        if self.training:
            z = self.sampling(z_mu, z_logvar)
        else:
            z = z_mu
        return z, z_mu, z_logvar


class VAE(nn.Module):
    def __init__(self, num_clusters, z_n=2, h1_n=16, h2_n=32, out_n=5):
        super(VAE, self).__init__()
        self.embed_mu = nn.Embedding(num_clusters, z_n)
        self.embed_logvar = nn.Embedding(num_clusters, z_n)
        self.fc1 = nn.Linear(z_n, h1_n)
        self.fc2 = nn.Linear(h1_n, h2_n)
        self.fc3 = nn.Linear(h2_n + z_n, out_n)

    def encode(self, x):
        x = x.long()
        mu = self.embed_mu(x)
        logvar = self.embed_logvar(x)
        return mu, logvar

    def sampling(self, mu, logvar):
        std = logvar.mul(0.5).exp_()
        eps = std.new_empty(std.size()).normal_()
        return eps.mul_(std).add_(mu)

    def decode(self, z):
        """
        Trying out skip connections, due to positive results reported here:
             https://adjidieng.github.io/Papers/skipvae.pdf
        """
        h1 = F.relu(self.fc1(z))
        h2 = F.relu(self.fc2(h1))
        h2_skip = torch.cat([h2, z], dim=1)
        return self.fc3(h2_skip)

    def forward(self, x):
        z_mu, z_logvar = self.encode(x)
        if self.training:
            z = self.sampling(z_mu, z_logvar)
        else:
            z = z_mu
        return self.decode(z), z_mu, z_logvar


class QueryDecoder(nn.Module):
    def __init__(self, n_z, num_queries, key_len, n_hidden1=30, n_hidden2=30):
        super().__init__()
        self.out_shape = (num_queries, key_len)
        self.fc1 = nn.Linear(n_z, n_hidden1)
        self.fc2 = nn.Linear(n_hidden1, n_hidden2)
        self.fc3 = nn.Linear(n_hidden2, num_queries * key_len)

    def forward(self, z):
        x = torch.tanh(self.fc1(z))
        x = torch.tanh(self.fc2(x))
        x = self.fc3(x)
        x = x.view(-1, *self.out_shape)
        return x


class MultiClusterModel2(nn.Module):
    LED_CHANNELS = 4
    NUM_CLUSTERS = 1

    def __init__(
        self, in_len, out_len, num_downsample, num_recordings, num_clusters
    ):
        super().__init__()
        self.in_len = in_len
        self.out_len = out_len
        self.num_input_channels = self.LED_CHANNELS + self.NUM_CLUSTERS
        # Linear
        self.linear_in_len = 1 + (in_len - 1) // (2**num_downsample)

        self.num_channels = [20, 100, 100, 100]
        self.downsample = [False, False, True, False]
        self.num_repeats = [None, 2, num_downsample - 1, 3]
        self.mid_kernel_size = 7
        self.kernel_sizes = [
            51,
            self.mid_kernel_size,
            self.mid_kernel_size,
            self.mid_kernel_size,
        ]
        self.expansion = 1
        # VAE
        self.z_dim = 2
        self.num_clusters = num_clusters
        self.num_embed = num_recordings * num_clusters
        self.n_h1 = 20
        self.n_h2 = 20
        # HyperResnet
        warehouse_size = 1500
        self.key_len = 16  # 32 = ~ sqrt(1000)

        # Huge memory store.
        self.warehouse = retinapy.nn.Conv1dWarehouse(
            max_in_channels=1,  # depth-wise conv.
            warehouse_size=warehouse_size,
            kernel_size=self.mid_kernel_size,
            key_len=self.key_len,
        )

        # Traditional conv layers.
        self.layer0 = nn.Sequential(
            nn.Conv1d(
                self.num_input_channels,
                self.num_channels[0],
                kernel_size=self.kernel_sizes[0],
                stride=2,
                padding=self.kernel_sizes[0] // 2,
                bias=True,
            ),
            nn.LeakyReLU(0.2, True),
            nn.Conv1d(
                self.num_channels[0],
                self.num_channels[0],
                kernel_size=self.kernel_sizes[0],
                stride=1,
                padding=self.kernel_sizes[0] // 2,
                bias=False,
            ),
            retinapy.nn.create_batch_norm(self.num_channels[0]),
            nn.LeakyReLU(0.2, True),
        )

        mid = []
        total_num_channels = 0
        for layer_idx in range(1, len(self.num_channels)):
            blocks = []
            for i in range(self.num_repeats[layer_idx]):
                num_in = (
                    self.num_channels[layer_idx - 1]
                    if (i == 0)
                    else self.num_channels[layer_idx]
                )
                res_block_F = retinapy.nn.ResBlock1d_F(
                    num_in,
                    self.num_channels[layer_idx] * self.expansion,
                    self.num_channels[layer_idx],
                    kernel_size=self.kernel_sizes[layer_idx],
                    downsample=self.downsample[layer_idx],
                )
                blocks.append(res_block_F)
                total_num_channels += res_block_F.num_channels()
            mid.append(nn.ModuleList(blocks))
        self.mid_layers = nn.ModuleList(mid)

        # Back to traditional conv layers.
        self.layer4 = nn.Conv1d(
            in_channels=self.num_channels[-1],
            out_channels=1,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
        )
        linear_in_len = 1 + (in_len - 1) // (2**num_downsample)
        self.linear = nn.Linear(
            in_features=linear_in_len,
            out_features=self.out_len,
        )
        # VAE
        self.vae = MiniVAE(num_recordings * num_clusters, z_n=self.z_dim)
        # VAE decode to attention query
        # 1 query per mid channel.
        self.query_decoder = QueryDecoder(
            self.z_dim, total_num_channels, self.key_len
        )

    def encode_vae(self, rec_id, cluster_id):
        id_ = rec_id * self.num_clusters + cluster_id
        z, z_mu, z_logvar = self.vae(id_)
        return z, z_mu, z_logvar

    def forward(self, snippet, rec_id, cluster_id):
        # Layer 0
        x = self.layer0(snippet)
        # VAE
        z, z_mu, z_logvar = self.encode_vae(rec_id, cluster_id)
        # Queries
        queries = self.query_decoder(z)
        weights_W, weights_b = self.warehouse(queries)
        idx_start = 0
        for module_list in self.mid_layers:
            for layer in module_list:
                x, idx_start = layer.forward(x, weights_W, weights_b, idx_start)
        x = self.layer4(x)
        x = self.linear(torch.flatten(x, start_dim=1))
        return x, z_mu, z_logvar


class CatMultiClusterModel(nn.Module):
    LED_CHANNELS = 4
    NUM_CLUSTERS = 1

    def __init__(
        self,
        in_len,
        out_len,
        num_downsample,
        num_recordings,
        num_clusters,
        z_dim=2,
    ):
        super().__init__()
        self.in_len = in_len
        self.out_len = out_len
        # led_channels, mean(led_channels), num_clusters, pos_encoding
        self.num_input_channels = self.LED_CHANNELS * 2 + self.NUM_CLUSTERS + 1
        # Linear
        self.linear_in_len = 1 + (in_len - 1) // (2**num_downsample)
        # L1
        self.l0_num_channels = 50
        # CNN parameters
        self.num_l1_blocks = num_downsample - 1
        self.num_l2_blocks = 3
        self.expansion = 2
        self.l1_num_channels = 100
        self.l2_num_channels = 200
        kernel_size = 21
        mid_kernel_size = 7
        # VAE
        self.z_dim = z_dim
        self.num_recordings = num_recordings
        self.num_clusters = num_clusters
        self.n_vae_out = 30

        self.layer0 = nn.Sequential(
            nn.Conv1d(
                self.num_input_channels,
                self.l0_num_channels,
                kernel_size=kernel_size,
                stride=2,
                padding=kernel_size // 2,
                bias=True,
            ),
            nn.LeakyReLU(0.2, True),
            nn.Conv1d(
                self.l0_num_channels,
                self.l0_num_channels,
                kernel_size=kernel_size,
                stride=1,
                padding=kernel_size // 2,
                bias=False,
            ),
            retinapy.nn.create_batch_norm(self.l0_num_channels),
            nn.LeakyReLU(0.2, True),
        )

        l1_blocks = []
        l1_num_in = self.l0_num_channels + self.n_vae_out
        for i in range(self.num_l1_blocks):
            num_in = self.l1_num_channels if i else l1_num_in
            l1_blocks.append(
                retinapy.nn.ResBlock1d(
                    num_in,
                    self.l1_num_channels * self.expansion,
                    self.l1_num_channels,
                    kernel_size=mid_kernel_size,
                    downsample=True,
                ),
            )
        self.layer1 = nn.Sequential(*l1_blocks)
        l2_blocks = []
        for i in range(self.num_l1_blocks):
            num_in = self.l2_num_channels if i else self.l1_num_channels
            l2_blocks.append(
                retinapy.nn.ResBlock1d(
                    num_in,
                    self.l2_num_channels * self.expansion,
                    self.l2_num_channels,
                    kernel_size=mid_kernel_size,
                    downsample=False,
                ),
            )
        self.layer2 = nn.Sequential(*l2_blocks)

        self.layer3 = nn.Conv1d(
            in_channels=self.l2_num_channels,
            out_channels=1,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
        )

        linear_in_len = 1 + (in_len - 1) // (2**num_downsample)
        self.linear = nn.Linear(
            in_features=linear_in_len,
            out_features=self.out_len,
        )
        # VAE
        self.vae = VAE(
            num_recordings * num_clusters, z_n=self.z_dim, out_n=self.n_vae_out
        )

    def encode_vae(self, rec_id, cluster_id):
        id_ = rec_id * self.num_clusters + cluster_id
        z_decode, z_mu, z_logvar = self.vae(id_)
        return z_decode, z_mu, z_logvar

    # def cat_downsample(self, x):
    #     ds = torchaudio.functional.lowpass_biquad(
    #             x[:,0:-1], sample_rate=992, cutoff_freq=10, Q=0.707
    #     )
    #     x = torch.cat([x, ds], dim=1)
    #     return x

    def cat_mean(self, snippet):
        m = (
            snippet[:, 0 : mea.NUM_STIMULUS_LEDS]
            .mean(dim=2, keepdim=True)
            .expand(-1, -1, snippet.shape[-1])
        )
        x = torch.cat([snippet, m], dim=1)
        return x

    def cat_z(self, x, z):
        z = z.unsqueeze(-1).expand(-1, -1, x.shape[-1])
        x = torch.cat([x, z], dim=1)
        return x

    def forward(self, snippet, rec_id, cluster_id):
        # VAE
        x = self.cat_mean(snippet)
        x = self.layer0(x)
        z_decode, z_mu, z_logvar = self.encode_vae(rec_id, cluster_id)
        x = self.cat_z(x, z_decode)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.linear(torch.flatten(x, start_dim=1))
        return x, z_mu, z_logvar


class MultiClusterModel(nn.Module):
    LED_CHANNELS = 4
    NUM_CLUSTERS = 1

    def __init__(
        self,
        in_len,
        out_len,
        num_downsample,
        num_recordings,
        num_clusters,
        z_dim=2,
    ):
        super(MultiClusterModel, self).__init__()
        self.in_len = in_len
        self.out_len = out_len
        # led_channels, mean(led_channels), num_clusters, pos_encoding
        self.num_input_channels = self.LED_CHANNELS * 2 + self.NUM_CLUSTERS + 1
        # Linear
        self.linear_in_len = 1 + (in_len - 1) // (2**num_downsample)
        # L1
        self.l0_num_channels = 50
        self.l1_num_channels = 50
        # CNN parameters
        self.num_l1_blocks = 1
        self.num_l2_blocks = num_downsample - 1
        self.num_l3_blocks = 1
        self.expansion = 1
        self.l2_num_channels = 100
        self.l3_num_channels = 200
        kernel_size = 21
        mid_kernel_size = 7
        # VAE
        self.z_dim = z_dim
        self.num_recordings = num_recordings
        self.num_clusters = num_clusters
        # HyperResnet
        warehouse_factor = 8
        self.hyper_hidden1 = 16
        self.hyper_hidden2 = 16

        self.layer0 = nn.Sequential(
            nn.Conv1d(
                self.num_input_channels,
                self.l0_num_channels,
                kernel_size=kernel_size,
                stride=2,
                padding=kernel_size // 2,
                bias=True,
            ),
            nn.LeakyReLU(0.2, True),
            nn.Conv1d(
                self.l0_num_channels,
                self.l0_num_channels,
                kernel_size=kernel_size,
                stride=1,
                padding=kernel_size // 2,
                bias=False,
            ),
            retinapy.nn.create_batch_norm(self.l0_num_channels),
            nn.LeakyReLU(0.2, True),
        )
        l1_blocks = []
        for i in range(self.num_l1_blocks):
            num_in = self.l1_num_channels if i else self.l0_num_channels
            hyper_res_block = retinapy.nn.HyperResBlock1d(
                num_in,
                self.l1_num_channels * self.expansion,
                self.l1_num_channels,
                warehouse_factor=warehouse_factor,
                kernel_size=mid_kernel_size,
                downsample=False,
            )
            hyper_decoder = retinapy.nn.HyperResDecoder(
                hyper_res_block,
                in_n=self.z_dim,
                hidden1_n=self.hyper_hidden1,
                hidden2_n=self.hyper_hidden2,
            )
            l1_blocks.append(hyper_decoder)
        self.layer1 = nn.ModuleList(l1_blocks)

        l2_blocks = []
        for i in range(self.num_l2_blocks):
            num_in = self.l2_num_channels if i else self.l1_num_channels
            l2_blocks.append(
                retinapy.nn.ResBlock1d(
                    num_in,
                    self.l2_num_channels * self.expansion,
                    self.l2_num_channels,
                    kernel_size=mid_kernel_size,
                    downsample=True,
                ),
            )
        self.layer2 = nn.Sequential(*l2_blocks)

        l3_blocks = []
        for i in range(self.num_l3_blocks):
            num_in = self.l3_num_channels if i else self.l2_num_channels
            hyper_res_block = retinapy.nn.HyperResBlock1d(
                num_in,
                self.l3_num_channels * self.expansion,
                self.l3_num_channels,
                warehouse_factor=warehouse_factor,
                kernel_size=mid_kernel_size,
                downsample=False,
            )
            hyper_decoder = retinapy.nn.HyperResDecoder(
                hyper_res_block,
                in_n=self.z_dim,
                hidden1_n=self.hyper_hidden1,
                hidden2_n=self.hyper_hidden2,
            )
            l3_blocks.append(hyper_decoder)
        self.layer3 = nn.ModuleList(l3_blocks)
        self.layer4 = nn.Conv1d(
            in_channels=self.l3_num_channels,
            out_channels=1,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
        )
        linear_in_len = 1 + (in_len - 1) // (2**num_downsample)
        self.linear = nn.Linear(
            in_features=linear_in_len,
            out_features=self.out_len,
        )
        # VAE
        self.vae = MiniVAE(num_recordings * num_clusters, z_n=self.z_dim)

    def encode_vae(self, rec_id, cluster_id):
        id_ = rec_id * self.num_clusters + cluster_id
        z, z_mu, z_logvar = self.vae(id_)
        return z, z_mu, z_logvar

    # def cat_downsample(self, x):
    #     ds = torchaudio.functional.lowpass_biquad(
    #             x[:,0:-1], sample_rate=992, cutoff_freq=10, Q=0.707
    #     )
    #     x = torch.cat([x, ds], dim=1)
    #     return x

    def cat_mean(self, snippet):
        m = (
            snippet[:, 0:-1]
            .mean(dim=2, keepdim=True)
            .expand(-1, -1, snippet.shape[-1])
        )
        x = torch.cat([snippet, m], dim=1)
        return x

    def forward(self, snippet, rec_id, cluster_id):
        x = self.cat_mean(snippet)
        x = self.layer0(x)
        # VAE
        z, z_mu, z_logvar = self.encode_vae(rec_id, cluster_id)
        # Hyper
        for l in self.layer1:
            x = l(x, z)
        x = self.layer2(x)
        for l in self.layer3:
            x = l(x, z)
        x = self.layer4(x)
        x = self.linear(torch.flatten(x, start_dim=1))
        return x, z_mu, z_logvar


class ClusteringTransformer(nn.Module):
    def __init__(
        self, in_len, out_len, stim_downsample, num_recordings, num_clusters,
        z_dim=2, num_heads=8, head_dim=64, num_tlayers=6):
        super().__init__()
        self.in_len = in_len
        self.num_recordings = num_recordings
        self.num_clusters = num_clusters
        self.z_dim = z_dim
        # Stimulus CNN
        k0_size = 21
        k1_size = 7
        expansion = 1
        self.l0a_num_channels = 10
        self.l0b_num_channels = 20
        self.l1_num_channels = 100
        self.cnn = nn.Sequential(
            nn.Conv1d(
                mea.NUM_STIMULUS_LEDS * 2,  # stimulus + mean(stimulus)
                self.l0a_num_channels,
                kernel_size=k0_size,
                stride=2,
                padding=(k0_size - 1) // 2,
                bias=True,
            ),
            nn.LeakyReLU(0.2, True),
            nn.Conv1d(
                self.l0a_num_channels,
                self.l0b_num_channels,
                kernel_size=k0_size,
                stride=1,
                padding=(k0_size - 1) // 2,
                bias=False,
            ),
            retinapy.nn.create_batch_norm(self.l0b_num_channels),
            nn.LeakyReLU(0.2, True),
            *[
                retinapy.nn.ResBlock1d(
                    self.l1_num_channels if i else self.l0b_num_channels,
                    self.l1_num_channels * expansion,
                    self.l1_num_channels,
                    kernel_size=k1_size,
                    downsample=True,
                )
                for i in range(stim_downsample - 1)
            ],
        )
        
        self.embed_dim = 128
        # VAE
        self.vae = VAE(
            self.num_recordings * self.num_clusters, 
            z_n=self.z_dim, 
            h1_n=32,
            h2_n=32,
            out_n=self.embed_dim)
        # Transformer
        self.stim_embed = nn.Conv1d(
            self.l1_num_channels, self.embed_dim, kernel_size=1
        )
        # Normally initialized nn.Parameter
        # 1092 = 992 + 100
        in_stim_len = self.in_len + out_len
        enc_stim_len = 1 + (in_stim_len - 1) // (2**stim_downsample)
        enc_len = enc_stim_len + 1 # 1 for VAE encoding.
        self.pos_embed = nn.Parameter(torch.randn(enc_len, self.embed_dim))

        mlp_expansion = 3
        self.transformer = retinapy.nn.Transformer(
            self.embed_dim,
            num_layers=num_tlayers,
            num_heads=num_heads,
            head_dim=head_dim,
            mlp_dim=self.embed_dim * mlp_expansion,
        )
        self.decode = nn.Sequential(
            nn.Linear(self.embed_dim, 1),
            einops.layers.torch.Rearrange("b c 1 -> b c"),
            nn.Linear(enc_len, out_len),
        )

    def cat_mean(self, stim):
        m = (
            stim[:, 0 : mea.NUM_STIMULUS_LEDS]
            .mean(dim=2, keepdim=True)
            .expand(-1, -1, stim.shape[-1])
        )
        x = torch.cat([stim, m], dim=1)
        return x

    def encode_stimulus(self, stim):
        x = self.cat_mean(stim)
        x = self.cnn(x)
        x = self.stim_embed(x)
        x = einops.rearrange(x, "b c l -> b l c")
        return x

    def encode_vae(self, rec_id, cluster_id):
        id_ = rec_id * self.num_clusters + cluster_id
        z_decode, z_mu, z_logvar = self.vae(id_)
        return z_decode, z_mu, z_logvar

    def forward(self, snippet, rec_id, cluster_id):
        z_decode, z_mu, z_logvar = self.encode_vae(rec_id, cluster_id)
        x_stim = self.encode_stimulus(snippet[:, 0:-1])
        z = einops.rearrange(z_decode, "b c -> b () c")
        x = torch.cat([x_stim, z], dim=1)
        x = x + self.pos_embed
        x = self.transformer(x)
        x = self.decode(x)
        return x, z_mu, z_logvar

class TransformerModel(nn.Module):
    def __init__(
        self, in_len, out_len, stim_downsample, num_recordings, num_clusters,
        z_dim=2, num_heads=8, head_dim=64, num_tlayers=6, spike_patch_len=8):
        super().__init__()
        self.in_len = in_len
        self.num_recordings = num_recordings
        self.num_clusters = num_clusters
        self.z_dim = z_dim
        # Stimulus CNN
        k0_size = 21
        k1_size = 7
        expansion = 1
        self.l0a_num_channels = 10
        self.l0b_num_channels = 20
        self.l1_num_channels = 100
        self.cnn = nn.Sequential(
            nn.Conv1d(
                mea.NUM_STIMULUS_LEDS * 2,  # stimulus + mean(stimulus)
                self.l0a_num_channels,
                kernel_size=k0_size,
                stride=2,
                padding=(k0_size - 1) // 2,
                bias=True,
            ),
            nn.LeakyReLU(0.2, True),
            nn.Conv1d(
                self.l0a_num_channels,
                self.l0b_num_channels,
                kernel_size=k0_size,
                stride=1,
                padding=(k0_size - 1) // 2,
                bias=False,
            ),
            retinapy.nn.create_batch_norm(self.l0b_num_channels),
            nn.LeakyReLU(0.2, True),
            *[
                retinapy.nn.ResBlock1d(
                    self.l1_num_channels if i else self.l0b_num_channels,
                    self.l1_num_channels * expansion,
                    self.l1_num_channels,
                    kernel_size=k1_size,
                    downsample=True,
                )
                for i in range(stim_downsample - 1)
            ],
        )
        
        self.embed_dim = 128
        # VAE
        self.vae = VAE(
            self.num_recordings * self.num_clusters, 
            z_n=self.z_dim, 
            h1_n=32,
            h2_n=32,
            out_n=self.embed_dim)
        # Transformer
        self.spike_patch_len = spike_patch_len
        self.stim_embed = nn.Conv1d(
            self.l1_num_channels, self.embed_dim, kernel_size=1
        )
        # Normally initialized nn.Parameter
        # 1092 = 992 + 100
        in_stim_len = self.in_len + out_len
        enc_stim_len = 1 + (in_stim_len - 1) // (2**stim_downsample)
        enc_spikes_len = math.ceil(self.in_len // self.spike_patch_len)
        self.spike_pad = enc_spikes_len * self.spike_patch_len - self.in_len
        enc_len = enc_stim_len + enc_spikes_len + 1 # 1 for VAE encoding.
        self.pos_embed = nn.Parameter(torch.randn(enc_len, self.embed_dim))

        self.spikes_embed = nn.Linear(self.spike_patch_len, self.embed_dim)
        mlp_expansion = 3
        self.transformer = retinapy.nn.Transformer(
            self.embed_dim,
            num_layers=num_tlayers,
            num_heads=num_heads,
            head_dim=head_dim,
            mlp_dim=self.embed_dim * mlp_expansion,
        )
        self.decode = nn.Sequential(
            nn.Linear(self.embed_dim, 1),
            einops.layers.torch.Rearrange("b c 1 -> b c"),
            nn.Linear(enc_len, out_len),
        )

    def cat_mean(self, stim):
        m = (
            stim[:, 0 : mea.NUM_STIMULUS_LEDS]
            .mean(dim=2, keepdim=True)
            .expand(-1, -1, stim.shape[-1])
        )
        x = torch.cat([stim, m], dim=1)
        return x

    def encode_stimulus(self, stim):
        x = self.cat_mean(stim)
        x = self.cnn(x)
        x = self.stim_embed(x)
        x = einops.rearrange(x, "b c l -> b l c")
        return x

    def encode_spikes(self, spikes):
        # Pad the spikes to be a multiple of the patch length.
        # Pad on the left, as the left has less of an effect on the output.
        x = F.pad(spikes, (self.spike_pad, 0))
        x = einops.rearrange(x, "b (t c) -> b t c", c=self.spike_patch_len)
        x = self.spikes_embed(x)
        return x

    def encode_vae(self, rec_id, cluster_id):
        id_ = rec_id * self.num_clusters + cluster_id
        z_decode, z_mu, z_logvar = self.vae(id_)
        return z_decode, z_mu, z_logvar

    def forward(self, snippet, rec_id, cluster_id):
        z_decode, z_mu, z_logvar = self.encode_vae(rec_id, cluster_id)
        x_stim = self.encode_stimulus(snippet[:, 0:-1])
        x_spikes = self.encode_spikes(snippet[:, -1][:, : self.in_len])
        z = einops.rearrange(z_decode, "b c -> b () c")
        x = torch.cat([x_stim, x_spikes, z], dim=1)
        x = x + self.pos_embed
        x = self.transformer(x)
        x = self.decode(x)
        return x, z_mu, z_logvar


class DistanceFieldCnnModel(nn.Module):
    LED_CHANNELS = 4
    NUM_CLUSTERS = 1

    def __init__(self, in_len, out_len):
        super(DistanceFieldCnnModel, self).__init__()
        self.in_len = in_len
        self.out_len = out_len
        # led_channels, mean(led_channels), num_clusters 
        self.num_input_channels = self.LED_CHANNELS * 2 + self.NUM_CLUSTERS
        self.l1_num_channels = 50
        self.l2_num_channels = 50
        self.l3_num_channels = 100
        kernel_size = 21
        mid_kernel_size = 7
        self.layer0 = nn.Sequential(
            nn.Conv1d(
                self.num_input_channels,
                self.l1_num_channels,
                kernel_size=kernel_size,
                stride=2,
                padding=(kernel_size - 1) // 2,
                bias=True,
            ),
            nn.LeakyReLU(0.2, True),
            nn.Conv1d(
                self.l1_num_channels,
                self.l1_num_channels,
                kernel_size=kernel_size,
                stride=1,
                padding=(kernel_size - 1) // 2,
                bias=False,
            ),
            retinapy.nn.create_batch_norm(self.l1_num_channels),
            nn.LeakyReLU(0.2, True),
        )
        self.layer1 = nn.Sequential(
            retinapy.nn.ResBlock1d(
                self.l1_num_channels,
                self.l2_num_channels,
                self.l2_num_channels,
                kernel_size=mid_kernel_size,
                downsample=False,
            ),
        )
        self.layer2_elements = []
        expansion = 1
        num_halves = 3  # There is 1 other downsamples other than the midlayers.
        for i in range(num_halves - 1):
            self.layer2_elements.append(
                retinapy.nn.ResBlock1d(
                    self.l2_num_channels,
                    self.l2_num_channels * expansion,
                    self.l2_num_channels,
                    kernel_size=mid_kernel_size,
                    downsample=True,
                )
            )
        self.layer2 = nn.Sequential(*self.layer2_elements)
        self.layer3 = retinapy.nn.ResBlock1d(
            self.l2_num_channels,
            self.l3_num_channels,
            self.l3_num_channels,
            kernel_size=mid_kernel_size,
            downsample=False,
        )
        self.layer4 = nn.Conv1d(
            in_channels=self.l3_num_channels,
            out_channels=1,
            kernel_size=5,
            stride=1,
            padding=2,
            bias=True,
        )
        linear_in_len = 1 + (in_len - 1) // 2**num_halves
        self.linear = nn.Linear(
            in_features=linear_in_len,
            out_features=self.out_len,
        )

    def cat_mean(self, snippet):
        m = (
            snippet[:, 0 : mea.NUM_STIMULUS_LEDS]
            .mean(dim=2, keepdim=True)
            .expand(-1, -1, snippet.shape[-1])
        )
        x = torch.cat([snippet, m], dim=1)
        return x

    def forward(self, x):
        x = self.cat_mean(x)
        x = self.layer0(x)
        # Keep or leave out? Interested to see a justification either way.
        # x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.linear(torch.flatten(x, start_dim=1))
        # Alternative parameter free head:
        # x = torch.squeeze(F.adaptive_avg_pool1d(x, output_size=self.out_len), dim=1)
        return x


class DistFieldToSpikeModel(nn.Module):
    def __init__(self, in_len):
        super(DistFieldToSpikeModel, self).__init__()
        self.act = nn.Softplus()
        k = 5
        c = 50
        n = 3
        self.layer1 = nn.Sequential(
            retinapy.nn.ResBlock1d(1, c, c, kernel_size=k, downsample=True),
            *[
                retinapy.nn.ResBlock1d(
                    c,
                    c,
                    c,
                    kernel_size=k,
                    downsample=True,
                )
                for _ in range(n - 1)
            ],
        )
        linear_in_len = (in_len // 2**n) * c
        self.linear = nn.Linear(in_features=linear_in_len, out_features=1)

    def forward(self, x):
        x = x.unsqueeze(dim=1)
        x = self.layer1(x)
        x = self.linear(torch.flatten(x, start_dim=1))
        x = self.act(x)
        x = torch.squeeze(x, dim=1)
        return x


class LinearNonlinear(nn.Module):
    # TODO: add options for 'rbf' and  'psp' activations.

    def __init__(self, in_n, out_n):
        super(LinearNonlinear, self).__init__()
        self.linear = nn.Linear(in_features=in_n, out_features=out_n)
        self.act = nn.Softplus()

    def forward(self, x):
        x = torch.flatten(x, start_dim=1)
        x = self.linear(x)
        x = self.act(x)
        # Remove the last dimension, as it is always 1.
        assert x.shape[1] == 1
        x = torch.squeeze(x, dim=1)
        return x


class DeepRetina2016(nn.Module):
    """
    From:
        https://github.com/baccuslab/deep-retina/blob/master/deepretina/models.py
    """

    def __init__(self, in_len, in_n, out_n):
        super(DeepRetina2016, self).__init__()
        self.in_n = in_n
        self.out_n = out_n
        self.in_len = in_len

        self.conv1 = nn.Conv2d(
            in_channels=self.in_n, out_channels=16, kernel_size=15
        )
        self.conv2 = nn.Conv2d(in_channels=16, out_channels=8, kernel_size=9)
        linear_in_len = in_len - 15 + 1 - 9 + 1
        self.linear = nn.Linear(linear_in_len, self.out_n)
        self.act = nn.Softplus()
        # self.linear = nn.Linear(in_features=?, out_features=self.num_softmax_classes)

    def forward(self, x):
        # TODO: add noise?
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = torch.flatten(x, start_dim=1)
        x = self.linear(x)
        x = self.act(x)
